from abc import ABC
from enum import unique, Enum


@unique
class ActionState(Enum):
    PRE_RUN = 0
    RUNNING = 1
    DONE = 2
    SOURCE_LIMIT_REACHED = 3
    SESSION_LIMIT_REACHED = 4


class ActionStatus(object):
    def __init__(self, state):
        self.state = state
        self.limit_state = None

    def set(self, state):
        self.state = state

    def get(self):
        return self.state

    def set_limit(self, limit_state):
        self.limit_state = limit_state

    def get_limit(self):
        return self.limit_state


class ActionsRunner(object):
    """An interface for actions-runner object"""

    ACTION_ID = "OVERRIDE"
    ACTION_ARGS = {"OVERRIDE": "OVERRIDE"}

    action_status = None

    def is_action_selected(self, args):
        raise NotImplementedError()

    def set_params(self, args):
        raise NotImplementedError()

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        raise NotImplementedError()


class CoreActionsRunner(ActionsRunner, ABC):
    """An interface for extra-actions-runner object"""



from insomniac.action_runners import *
from insomniac.safely_runner import run_safely
from insomniac.utils import *


class UnfollowActionRunner(CoreActionsRunner):
    ACTION_ID = "unfollow"
    ACTION_ARGS = {
        "unfollow": {
            "help": 'unfollow at most given number of users. Only users followed by this script will '
                    'be unfollowed. The order is from oldest to newest followings. '
                    'It can be a number (e.g. 100) or a range (e.g. 100-200)',
            "metavar": '100-200'
        },
        "unfollow_followed_by_anyone": {
            'help': 'By default, only profiles that been followed by the bot will be unfollowed. Set this parameter '
                    'if you want to unfollow any profile (even if not been followed by the bot)',
            'action': 'store_true'
        },
        "unfollow_non_followers": {
            'help': 'unfollow only profiles that are not following you',
            'action': 'store_true'
        }
    }

    unfollow_followed_by_anyone = False
    unfollow_non_followers = False

    def is_action_selected(self, args):
        return args.unfollow is not None

    def set_params(self, args):
        if args.unfollow_followed_by_anyone is not None:
            self.unfollow_followed_by_anyone = True

        if args.unfollow_non_followers is not None:
            self.unfollow_non_followers = True

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        from insomniac.action_runners.unfollow.action_unfollow import unfollow, get_unfollow_restriction

        unfollow_restriction = get_unfollow_restriction(self.unfollow_followed_by_anyone, self.unfollow_non_followers)

        self.action_status = ActionStatus(ActionState.PRE_RUN)

        @run_safely(device_wrapper=device_wrapper)
        def job():
            self.action_status.set(ActionState.RUNNING)
            unfollow(device=device_wrapper.get(),
                     on_action=on_action,
                     storage=storage,
                     unfollow_restriction=unfollow_restriction,
                     session_state=session_state,
                     is_limit_reached=is_limit_reached,
                     action_status=self.action_status)
            print("Unfollowed " + str(session_state.totalUnfollowed) + " " + unfollow_restriction.value + ", finish.")
            self.action_status.set(ActionState.DONE)

        while not self.action_status.get() == ActionState.DONE:
            job()



from insomniac.action_runners.core import ActionsRunner, CoreActionsRunner, ActionStatus, ActionState
from insomniac.action_runners.interact import InteractBySourceActionRunner, InteractByTargetsActionRunner
from insomniac.action_runners.unfollow import UnfollowActionRunner


def get_core_action_runners_classes():
    return CoreActionsRunner.__subclasses__()



from abc import ABC
from enum import unique, Enum

from insomniac.action_runners.actions_runners_manager import ActionState
from insomniac.actions_types import LikeAction, InteractAction, FollowAction, UnfollowAction, StoryWatchAction, \
    GetProfileAction
from insomniac.utils import *


@unique
class LimitType(Enum):
    SESSION = 0
    SOURCE = 1


class LimitsManager(object):
    """This class manages the actions-limits of an insomniac-session"""

    limits = {
        str(LimitType.SOURCE): {},
        str(LimitType.SESSION): {}
    }

    def __init__(self):
        for clazz in get_core_limits_classes():
            instance = clazz()
            self.limits[str(instance.LIMIT_TYPE)][instance.LIMIT_ID] = instance

    def get_limits_args(self):
        limit_args = {}

        for limits_type, limits in self.limits.items():
            for limit_id, limit in limits.items():
                for arg, info in limit.LIMIT_ARGS.items():
                    limit_args.update({arg: info})

        return limit_args

    def set_limits(self, args):
        for limits_type, limits in self.limits.items():
            for limit_id, limit in limits.items():
                limit.set_limit(args)

    def update_state(self, action):
        for limits_type, limits in self.limits.items():
            for limit_id, limit in limits.items():
                limit.update_state(action)

    def is_limit_reached_for_action(self, action, session_state):
        """
        :return: three values: whether limit was reached,
        LIMIT_ID of source limit if the reached limit is LimitType.SOURCE or None otherwise,
        LIMIT_ID of session limit if the reached limit is LimitType.SESSION or None otherwise
        """

        reached_source_limit = None
        reached_session_limit = None
        is_limit_reached = False

        for limit_id, limit in self.limits[str(LimitType.SOURCE)].items():
            if limit.is_reached_for_action(action, session_state):
                reached_source_limit = limit.LIMIT_ID
                is_limit_reached = True
                break

        for limit_id, limit in self.limits[str(LimitType.SESSION)].items():
            if limit.is_reached_for_action(action, session_state):
                reached_session_limit = limit.LIMIT_ID
                is_limit_reached = True
                break

        return is_limit_reached, reached_source_limit, reached_session_limit


class Limit(object):
    """An interface for limit object"""

    LIMIT_ID = "OVERRIDE"
    LIMIT_TYPE = "OVERRIDE"
    LIMIT_ARGS = {"OVERRIDE": "OVERRIDE"}

    def set_limit(self, args):
        raise NotImplementedError()

    def update_state(self, action):
        raise NotImplementedError()

    def is_reached_for_action(self, action, session_state):
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()


class CoreLimit(Limit, ABC):
    """An interface for extra-limit object"""


class TotalLikesLimit(CoreLimit):
    LIMIT_ID = "total_likes_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_likes_limit": {
            "help": "limit on total amount of likes during the session, 300 by default. "
                    "It can be a number presenting specific limit (e.g. 300) or a range (e.g. 100-120)",
            "metavar": "300",
            "default": "1000"
        }
    }

    total_likes_limit = 1000

    def set_limit(self, args):
        if args.total_likes_limit is not None:
            self.total_likes_limit = get_value(args.total_likes_limit, "Total likes limit: {}", 1000)

    def is_reached_for_action(self, action, session_state):
        if not type(action) == LikeAction:
            return False

        return session_state.totalLikes >= self.total_likes_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class TotalInteractionsLimit(CoreLimit):
    LIMIT_ID = "total_interactions_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_interactions_limit": {
            "help": "number of total interactions (successful & unsuccessful) per session, disabled by default. "
                    "It can be a number (e.g. 70) or a range (e.g. 60-80)",
            "metavar": '60-80'
        }
    }

    total_interactions_limit = None

    def set_limit(self, args):
        if args.total_interactions_limit is not None:
            self.total_interactions_limit = get_value(args.total_interactions_limit, "Total interactions limit: {}", 1000)

    def is_reached_for_action(self, action, session_state):
        if self.total_interactions_limit is None:
            return False

        if not type(action) == InteractAction:
            return False

        return sum(session_state.totalInteractions.values()) >= self.total_interactions_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class TotalSuccessfulInteractionsLimit(CoreLimit):
    LIMIT_ID = "total_successful_interactions_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_successful_interactions_limit": {
            "help": "number of total successful interactions per session, disabled by default. "
                    "It can be a number (e.g. 70) or a range (e.g. 60-80)",
            "metavar": '60-80'
        }
    }

    total_successful_interactions_limit = None

    def set_limit(self, args):
        if args.total_successful_interactions_limit is not None:
            self.total_successful_interactions_limit = get_value(args.total_successful_interactions_limit,
                                                                 "Total successful-interactions limit: {}", 1000)

    def is_reached_for_action(self, action, session_state):
        if self.total_successful_interactions_limit is None:
            return False

        if not type(action) == InteractAction:
            return False

        return sum(session_state.successfulInteractions.values()) >= self.total_successful_interactions_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class TotalFollowLimit(CoreLimit):
    LIMIT_ID = "total_follow_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_follow_limit": {
            "help": "limit on total amount of follows during the session, disabled by default. "
                    "It can be a number (e.g. 27) or a range (e.g. 20-30)",
            "metavar": "50"
        }
    }

    total_follow_limit = None

    def set_limit(self, args):
        if args.total_follow_limit is not None:
            self.total_follow_limit = get_value(args.total_follow_limit, "Total follow limit: {}", 70)

    def is_reached_for_action(self, action, session_state):
        if self.total_follow_limit is None:
            return False
        
        if not type(action) == FollowAction:
            return False

        return sum(session_state.totalFollowed.values()) >= self.total_follow_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class TotalStoryWatchLimit(CoreLimit):
    LIMIT_ID = "total_story_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_story_limit": {
            "help": "limit on total amount of stories watches during the session, disabled by default. "
                    "It can be a number (e.g. 27) or a range (e.g. 20-30)",
            "metavar": "300",
        }
    }

    total_story_limit = None

    def set_limit(self, args):
        if args.total_story_limit is not None:
            self.total_story_limit = get_value(args.total_story_limit, "Total story-watches limit: {}", 1000)

    def is_reached_for_action(self, action, session_state):
        if self.total_story_limit is None:
            return False
        
        if not type(action) == StoryWatchAction:
            return False

        return session_state.totalStoriesWatched >= self.total_story_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class SourceInteractionsLimit(CoreLimit):
    LIMIT_ID = "interactions_count"
    LIMIT_TYPE = LimitType.SOURCE
    LIMIT_ARGS = {
        "interactions_count": {
            "help": "number of interactions per each blogger/hashtag, 70 by default. "
                    "It can be a number (e.g. 70) or a range (e.g. 60-80). "
                    "Only successful interactions count",
            "metavar": "40",
            "default": "70"
        }
    }

    interactions_count = 70

    def set_limit(self, args):
        if args.interactions_count is not None:
            self.interactions_count = get_value(args.interactions_count, "Interactions count: {}", 70)

    def is_reached_for_action(self, action, session_state):
        if not type(action) == InteractAction:
            return False

        successful_interactions_count = session_state.successfulInteractions.get(action.source)

        return successful_interactions_count and successful_interactions_count >= self.interactions_count

    def reset(self):
        pass

    def update_state(self, action):
        pass


class SourceFollowLimit(CoreLimit):
    LIMIT_ID = "follow_limit"
    LIMIT_TYPE = LimitType.SOURCE
    LIMIT_ARGS = {
        "follow_limit": {
            "help": "limit on amount of follows during interaction with each one user's followers, "
                    "disabled by default. It can be a number (e.g. 10) or a range (e.g. 6-9)",
            "metavar": "7-8",
        }
    }

    follow_limit = None

    def set_limit(self, args):
        if args.follow_limit is not None:
            self.follow_limit = get_value(args.follow_limit, "Follow limit: {}", 10)

    def is_reached_for_action(self, action, session_state):
        if self.follow_limit is None:
            return False

        if not type(action) == FollowAction:
            return False

        followed_count = session_state.totalFollowed.get(action.source)
        return followed_count is not None and followed_count >= self.follow_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class UnfollowingLimit(CoreLimit):
    LIMIT_ID = "unfollowing_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {}

    unfollow_limit = None

    def set_limit(self, args):
        if args.unfollow is not None:
            self.unfollow_limit = get_value(args.unfollow, "Unfollow: {}", 100)

    def is_reached_for_action(self, action, session_state):
        if self.unfollow_limit is None:
            return False

        if not type(action) == UnfollowAction:
            return False

        return session_state.totalUnfollowed >= self.unfollow_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class MinFollowing(CoreLimit):
    LIMIT_ID = "min_following"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "min_following": {
            "help": 'minimum amount of followings, after reaching this amount unfollow stops',
            "metavar": '100',
            "default": "0"
        }
    }

    min_following_limit = 0

    def set_limit(self, args):
        self.min_following_limit = int(args.min_following)

    def is_reached_for_action(self, action, session_state):
        if not type(action) == UnfollowAction:
            return False

        initial_following = session_state.my_following_count
        unfollowed_count = session_state.totalUnfollowed

        return initial_following - unfollowed_count <= self.min_following_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class MaxFollowing(CoreLimit):
    LIMIT_ID = "max_following"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "max_following": {
            "help": 'maximum amount of followings, after reaching this amount follow stops. disabled by default',
            "metavar": '100'
        }
    }

    max_following_limit = None

    def set_limit(self, args):
        if args.max_following is not None:
            self.max_following_limit = int(args.max_following)

    def is_reached_for_action(self, action, session_state):
        if self.max_following_limit is None:
            return False

        if not type(action) == FollowAction:
            return False

        initial_following = session_state.my_following_count
        followed_count = sum(session_state.totalFollowed.values())

        return initial_following + followed_count >= self.max_following_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class TotalGetProfileLimit(CoreLimit):
    LIMIT_ID = "total_get_profile_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "total_get_profile_limit": {
            "help": "limit on total amount of get-profile actions during the session, disabled by default. "
                    "It can be a number (e.g. 600) or a range (e.g. 500-700)",
            "metavar": "1500"
        }
    }

    total_get_profile_limit = None

    def set_limit(self, args):
        if args.total_get_profile_limit is not None:
            self.total_get_profile_limit = get_value(args.total_get_profile_limit, "Total get-profile limit: {}", 1000)

    def is_reached_for_action(self, action, session_state):
        if self.total_get_profile_limit is None:
            return False

        if not type(action) == GetProfileAction:
            return False

        return session_state.totalGetProfile >= self.total_get_profile_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


class SessionTimeMaxLengthLimit(CoreLimit):
    LIMIT_ID = "session_length_in_mins_limit"
    LIMIT_TYPE = LimitType.SESSION
    LIMIT_ARGS = {
        "session_length_in_mins_limit": {
            "help": "limit the session length by time (minutes), disabled by default. "
                    "It can be a number (e.g. 60) or a range (e.g. 40-70)",
            "metavar": "50-60"
        }
    }

    session_length_in_mins_limit = None

    def set_limit(self, args):
        if args.session_length_in_mins_limit is not None:
            self.session_length_in_mins_limit = get_value(args.session_length_in_mins_limit, "Session max-length (minutes): {}", 60)

    def is_reached_for_action(self, action, session_state):
        if self.session_length_in_mins_limit is None:
            return False

        # Apply this limit for every action (no action-type condition)

        delta = datetime.now() - session_state.startTime
        mins_delta = delta.seconds // 60

        return mins_delta > self.session_length_in_mins_limit

    def reset(self):
        pass

    def update_state(self, action):
        pass


def get_core_limits_classes():
    return CoreLimit.__subclasses__()


def process_limits(is_limit_reached, reached_session_limit, reached_source_limit, action_status, limit_display_name):
    if is_limit_reached:
        # Reached session limit, stop the action
        if reached_session_limit is not None:
            print(COLOR_OKBLUE + "{0} session-limit {1} has been reached. Stopping activity."
                  .format(limit_display_name, reached_session_limit) + COLOR_ENDC)
            action_status.set_limit(ActionState.SESSION_LIMIT_REACHED)
        else:
            print(COLOR_OKBLUE + "{0} source-limit {1} has been reached. Moving to next source."
                  .format(limit_display_name, reached_source_limit) + COLOR_ENDC)
            action_status.set_limit(ActionState.SOURCE_LIMIT_REACHED)

        return False

    return True



import random

from insomniac.action_runners import *
from insomniac.safely_runner import run_safely
from insomniac.utils import *


class InteractBySourceActionRunner(CoreActionsRunner):
    ACTION_ID = "interact"
    ACTION_ARGS = {
        "likes_count": {
            "help": "number of likes for each interacted user, 2 by default. "
                    "It can be a number (e.g. 2) or a range (e.g. 2-4)",
            'metavar': '2-4',
            "default": '2'
        },
        "like_percentage": {
            "help": "likes given percentage of interacted users, 100 by default",
            "metavar": '50',
            "default": '100'
        },
        "follow_percentage": {
            "help": "follow given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "interact": {
            "nargs": '+',
            "help": 'list of hashtags and usernames. Usernames should start with \"@\" symbol. '
                    'You can specify the way of interaction after a \"-\" sign: @username-followers, '
                    '@username-following, hashtag-top-likers, hashtag-recent-likers',
            "default": [],
            "metavar": ('hashtag-top-likers', '@username-followers')
        },
        "interaction_users_amount": {
            "help": 'add this argument to select an amount of users from the interact-list '
                    '(users are randomized). It can be a number (e.g. 4) or a range (e.g. 3-8)',
            'metavar': '3-8'
        },
        "stories_count": {
            "help": 'number of stories to watch for each user, disabled by default. '
                    'It can be a number (e.g. 2) or a range (e.g. 2-4)',
            'metavar': '3-8'
        },
    }

    likes_count = '2'
    follow_percentage = 0
    like_percentage = 100
    interact = []
    stories_count = '0'

    def is_action_selected(self, args):
        return args.interact is not None and len(args.interact) > 0

    def set_params(self, args):
        if args.likes_count is not None:
            self.likes_count = args.likes_count

        if args.stories_count is not None:
            self.stories_count = args.stories_count

        if args.interact is not None:
            self.interact = args.interact.copy()
            self.interact = [source if source[0] == '@' else ('#' + source) for source in self.interact]

        if args.follow_percentage is not None:
            self.follow_percentage = int(args.follow_percentage)

        if args.like_percentage is not None:
            self.like_percentage = int(args.like_percentage)

        if args.interaction_users_amount is not None:
            if len(self.interact) > 0:
                users_amount = get_value(args.interaction_users_amount, "Interaction user amount {}", 100)

                if users_amount >= len(self.interact):
                    print("interaction-users-amount parameter is equal or higher then the users-interact list. "
                          "Choosing all list for interaction.")
                else:
                    amount_to_remove = len(self.interact) - users_amount
                    for i in range(0, amount_to_remove):
                        self.interact.remove(random.choice(self.interact))

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        from insomniac.action_runners.interact.action_handle_blogger import handle_blogger, extract_blogger_instructions
        from insomniac.action_runners.interact.action_handle_hashtag import handle_hashtag, extract_hashtag_instructions

        random.shuffle(self.interact)

        for source in self.interact:
            self.action_status = ActionStatus(ActionState.PRE_RUN)

            if source[0] == '@':
                is_myself = source[1:] == session_state.my_username
                print_timeless("")
                print(COLOR_BOLD + "Handle " + source + (is_myself and " (it\'s you)" or "") + COLOR_ENDC)
            elif source[0] == '#':
                print_timeless("")
                print(COLOR_BOLD + "Handle " + source + COLOR_ENDC)

            @run_safely(device_wrapper=device_wrapper)
            def job():
                self.action_status.set(ActionState.RUNNING)
                if source[0] == '@':
                    source_name, instructions = extract_blogger_instructions(source)
                    handle_blogger(device_wrapper.get(),
                                   source_name[1:],  # drop "@"
                                   instructions,
                                   session_state,
                                   self.likes_count,
                                   self.stories_count,
                                   self.follow_percentage,
                                   self.like_percentage,
                                   storage,
                                   on_action,
                                   is_limit_reached,
                                   is_passed_filters,
                                   self.action_status)
                elif source[0] == '#':
                    source_name, instructions = extract_hashtag_instructions(source)
                    handle_hashtag(device_wrapper.get(),
                                   source_name[1:],  # drop "#"
                                   instructions,
                                   session_state,
                                   self.likes_count,
                                   self.stories_count,
                                   self.follow_percentage,
                                   self.like_percentage,
                                   storage,
                                   on_action,
                                   is_limit_reached,
                                   is_passed_filters,
                                   self.action_status)

                self.action_status.set(ActionState.DONE)

            while not self.action_status.get() == ActionState.DONE:
                job()
                if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED or \
                   self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                    break

            if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED:
                continue

            if self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                break


class InteractByTargetsActionRunner(CoreActionsRunner):
    ACTION_ID = "interact_targets"
    ACTION_ARGS = {
        "interact_targets": {
            "help": "use this argument in order to interact with profiles from targets.txt",
            'metavar': 'True / False'
        },
        "likes_count": {
            "help": "number of likes for each interacted user, 2 by default. "
                    "It can be a number (e.g. 2) or a range (e.g. 2-4)",
            'metavar': '2-4',
            "default": '2'
        },
        "follow_percentage": {
            "help": "follow given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "like_percentage": {
            "help": "likes given percentage of interacted users, 100 by default",
            "metavar": '50',
            "default": '100'
        },
        "stories_count": {
            "help": 'number of stories to watch for each user, disabled by default. '
                    'It can be a number (e.g. 2) or a range (e.g. 2-4)',
            'metavar': '3-8'
        }
    }

    likes_count = '2'
    follow_percentage = 0
    like_percentage = 100
    stories_count = '0'

    def is_action_selected(self, args):
        return args.interact_targets is not None

    def set_params(self, args):
        if args.likes_count is not None:
            self.likes_count = args.likes_count

        if args.stories_count is not None:
            self.stories_count = args.stories_count

        if args.follow_percentage is not None:
            self.follow_percentage = int(args.follow_percentage)

        if args.like_percentage is not None:
            self.like_percentage = int(args.like_percentage)

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        from insomniac.action_runners.interact.action_handle_target import handle_target

        target = storage.get_target()
        while target is not None:
            self.action_status = ActionStatus(ActionState.PRE_RUN)

            print_timeless("")
            print(COLOR_BOLD + "Handle @" + target + COLOR_ENDC)

            @run_safely(device_wrapper=device_wrapper)
            def job():
                self.action_status.set(ActionState.RUNNING)
                handle_target(device_wrapper.get(),
                              target,
                              session_state,
                              self.likes_count,
                              self.stories_count,
                              self.follow_percentage,
                              self.like_percentage,
                              storage,
                              on_action,
                              is_limit_reached,
                              is_passed_filters,
                              self.action_status)

                self.action_status.set(ActionState.DONE)

            while not self.action_status.get() == ActionState.DONE:
                job()
                if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED or \
                        self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                    break

            if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED:
                continue

            if self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                break

            target = storage.get_target()



from enum import Enum, unique
from random import uniform
from re import search

from insomniac.utils import *

# How long we're waiting until UI element appears (loading content + animation)
UI_TIMEOUT_LONG = 5
UI_TIMEOUT_SHORT = 1


def create_device(is_old, device_id, app_id):
    print("Using uiautomator v" + ("1" if is_old else "2"))
    try:
        return DeviceFacade(is_old, device_id, app_id)
    except ImportError as e:
        print(COLOR_FAIL + str(e) + COLOR_ENDC)
        return None


class DeviceFacade:
    deviceV1 = None  # uiautomator
    deviceV2 = None  # uiautomator2
    width = None
    height = None
    device_id = None
    app_id = None

    def __init__(self, is_old, device_id, app_id):
        self.device_id = device_id
        self.app_id = app_id
        if is_old:
            try:
                import uiautomator
                self.deviceV1 = uiautomator.device if device_id is None else uiautomator.Device(device_id)
            except ImportError:
                raise ImportError("Please install uiautomator: pip3 install uiautomator")
        else:
            try:
                import uiautomator2
                self.deviceV2 = uiautomator2.connect() if device_id is None else uiautomator2.connect(device_id)
            except ImportError:
                raise ImportError("Please install uiautomator2: pip3 install uiautomator2")

    def is_old(self):
        return self.deviceV1 is not None

    def find(self, *args, **kwargs):
        if self.deviceV1 is not None:
            import uiautomator
            try:
                view = self.deviceV1(*args, **kwargs)
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
        else:
            import uiautomator2
            try:
                view = self.deviceV2(*args, **kwargs)
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

    def back(self):
        if self.deviceV1 is not None:
            self.deviceV1.press.back()
        else:
            self.deviceV2.press("back")

    def screen_click(self, place):
        w, h = self._get_screen_size()
        if place == DeviceFacade.Place.RIGHT:
            left = int(w * 3 / 4)
            top = int(h / 2)
        else:
            return

        if self.deviceV1 is not None:
            self.deviceV1.click(left, top)
        else:
            self.deviceV2.click(left, top)

    def screenshot(self, path):
        if self.deviceV1 is not None:
            self.deviceV1.screenshot(path)
        else:
            self.deviceV2.screenshot(path)

    def dump_hierarchy(self, path):
        if self.deviceV1 is not None:
            xml_dump = self.deviceV1.dump()
        else:
            xml_dump = self.deviceV2.dump_hierarchy()

        with open(path, 'w', encoding="utf-8") as outfile:
            outfile.write(xml_dump)

    def is_screen_on(self):
        return self.get_info()["screenOn"]

    def press_power(self):
        if self.deviceV1 is not None:
            self.deviceV1.press.power()
        else:
            self.deviceV2.press("power")

    def is_screen_locked(self):
        cmd = f"adb {'' if self.device_id is None else ('-s '+ self.device_id)} shell dumpsys window"

        cmd_res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True, encoding="utf8")
        data = cmd_res.stdout.strip()
        flag = search("mDreamingLockscreen=(true|false)", data)
        return True if flag.group(1) == "true" else False

    def is_alive(self):
        if self.deviceV1 is not None:
            return self.deviceV1.alive()
        else:
            return self.deviceV2._is_alive()

    def wake_up(self):
        """ Make sure agent is alive or bring it back up before starting. """
        attempts = 0
        while not self.is_alive() and attempts < 5:
            self.get_info()
            attempts += 1

    def unlock(self):
        self.swipe(DeviceFacade.Direction.TOP, 0.8)
        if self.is_screen_locked():
            self.swipe(DeviceFacade.Direction.RIGHT, 0.8)

    def screen_off(self):
        if self.deviceV1 is not None:
            self.deviceV1.screen.off()
        else:
            self.deviceV2.screen_off()

    def swipe(self, direction: "DeviceFacade.Direction", scale=0.5):
        """Swipe finger in the `direction`.
        Scale is the sliding distance. Default to 50% of the screen width
        """

        if self.deviceV1 is not None:
            def _swipe(_from, _to):
                self.deviceV1.swipe(_from[0], _from[1], _to[0], _to[1])

            lx, ly = 0, 0
            rx, ry = self._get_screen_size()

            width, height = rx - lx, ry - ly

            h_offset = int(width * (1 - scale)) // 2
            v_offset = int(height * (1 - scale)) // 2

            left = lx + h_offset, ly + height // 2
            up = lx + width // 2, ly + v_offset
            right = rx - h_offset, ly + height // 2
            bottom = lx + width // 2, ry - v_offset

            if direction == DeviceFacade.Direction.TOP:
                _swipe(bottom, up)
            elif direction == DeviceFacade.Direction.RIGHT:
                _swipe(left, right)
            elif direction == DeviceFacade.Direction.LEFT:
                _swipe(right, left)
            elif direction == DeviceFacade.Direction.BOTTOM:
                _swipe(up, bottom)
        else:
            swipe_dir = ""
            if direction == DeviceFacade.Direction.TOP:
                swipe_dir = "up"
            elif direction == DeviceFacade.Direction.RIGHT:
                swipe_dir = "right"
            elif direction == DeviceFacade.Direction.LEFT:
                swipe_dir = "left"
            elif direction == DeviceFacade.Direction.BOTTOM:
                swipe_dir = "down"
            self.deviceV2.swipe_ext(swipe_dir, scale=scale)

    def swipe_points(self, sx, sy, ex, ey):
        if self.deviceV1 is not None:
            import uiautomator
            try:
                self.deviceV1.swipePoints([[sx, sy], [ex, ey]])
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
        else:
            import uiautomator2
            try:
                self.deviceV2.swipe_points([[sx, sy], [ex, ey]], uniform(0.2, 0.6))
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)

    def get_info(self):
        if self.deviceV1 is not None:
            return self.deviceV1.info
        else:
            return self.deviceV2.info

    def is_keyboard_open(self):
        cmd = f"adb {'' if self.device_id is None else ('-s '+ self.device_id)} shell dumpsys input_method"

        cmd_res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True, encoding="utf8")
        data = cmd_res.stdout.strip()
        flag = search("mInputShown=(true|false)", data)
        return True if flag.group(1) == "true" else False

    def close_keyboard(self):
        print("Closing keyboard...")
        if self.is_keyboard_open():
            print("Keyboard is open, closing it by pressing back")
            self.back()
            print("Verifying again that keyboard is closed")
            if self.is_keyboard_open():
                print(COLOR_FAIL + "Keyboard is open and couldn't be closed for some reason" + COLOR_ENDC)
            else:
                print("The device keyboard is closed now.")

            return

        print("The device keyboard is already closed.")

    def _get_screen_size(self):
        if self.width is not None and self.height is not None:
            return self.width, self.height

        if self.deviceV1 is not None:
            self.width = self.deviceV1.info['displayWidth']
            self.height = self.deviceV1.info['displayHeight']
        else:
            self.width = self.deviceV2.info['displayWidth']
            self.height = self.deviceV2.info['displayHeight']

        return self.width, self.height

    class View:
        deviceV1 = None  # uiautomator
        viewV1 = None  # uiautomator
        deviceV2 = None  # uiautomator2
        viewV2 = None  # uiautomator2

        def __init__(self, is_old, view, device):
            if is_old:
                self.viewV1 = view
                self.deviceV1 = device
            else:
                self.viewV2 = view
                self.deviceV2 = device

        def __iter__(self):
            children = []
            if self.viewV1 is not None:
                import uiautomator
                try:
                    for item in self.viewV1:
                        children.append(DeviceFacade.View(is_old=True, view=item, device=self.deviceV1))
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    for item in self.viewV2:
                        children.append(DeviceFacade.View(is_old=False, view=item, device=self.deviceV2))
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            return iter(children)

        def child(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.child(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.child(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

        def right(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.right(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.right(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

        def left(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.left(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.left(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

        def up(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.up(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.up(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

        def down(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.down(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.deviceV1)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.down(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.deviceV2)

        def click(self, mode=None):
            mode = DeviceFacade.Place.WHOLE if mode is None else mode
            if mode == DeviceFacade.Place.WHOLE:
                x_offset = uniform(0.15, 0.85)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.LEFT:
                x_offset = uniform(0.15, 0.4)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.CENTER:
                x_offset = uniform(0.4, 0.6)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.RIGHT:
                x_offset = uniform(0.6, 0.85)
                y_offset = uniform(0.15, 0.85)

            else:
                x_offset = 0.5
                y_offset = 0.5

            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.click.wait()
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    self.viewV2.click(UI_TIMEOUT_LONG, offset=(x_offset, y_offset))
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def double_click(self, padding=0.3):
            """
            Double click randomly in the selected view using padding
            padding: % of how far from the borders we want the double click to happen.
            """

            if self.viewV1 is not None:
                self._double_click_v1()
            else:
                self._double_click_v2(padding)

        def scroll(self, direction):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV1.scroll.toBeginning(max_swipes=1)
                    else:
                        self.viewV1.scroll.toEnd(max_swipes=1)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV2.scroll.toBeginning(max_swipes=1)
                    else:
                        self.viewV2.scroll.toEnd(max_swipes=1)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def swipe(self, direction):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV1.fling.toBeginning(max_swipes=5)
                    else:
                        self.viewV1.fling.toEnd(max_swipes=5)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV2.fling.toBeginning(max_swipes=5)
                    else:
                        self.viewV2.fling.toEnd(max_swipes=5)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def exists(self, quick=False):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    return self.viewV1.exists
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.exists(UI_TIMEOUT_SHORT if quick else UI_TIMEOUT_LONG)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def wait(self):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.deviceV1.wait.idle()
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return True
            else:
                import uiautomator2
                try:
                    return self.viewV2.wait(timeout=UI_TIMEOUT_LONG)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def get_bounds(self):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    return self.viewV1.bounds
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.info['bounds']
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def get_text(self, retry=True):
            max_attempts = 1 if not retry else 3
            attempts = 0

            if self.viewV1 is not None:
                import uiautomator
                while attempts < max_attempts:
                    attempts += 1
                    try:
                        text = self.viewV1.text
                        if text is None:
                            if attempts < max_attempts:
                                print(COLOR_REPORT + "Could not get text. Waiting 2 seconds and trying again..." + COLOR_ENDC)
                                sleep(2)  # wait 2 seconds and retry
                                continue
                        else:
                            return text
                    except uiautomator.JsonRPCError as e:
                        raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                while attempts < max_attempts:
                    attempts += 1
                    try:
                        text = self.viewV2.info['text']
                        if text is None:
                            if attempts < max_attempts:
                                print(COLOR_REPORT + "Could not get text. "
                                                     "Waiting 2 seconds and trying again..." + COLOR_ENDC)
                                sleep(2)  # wait 2 seconds and retry
                                continue
                        else:
                            return text
                    except uiautomator2.JSONRPCError as e:
                        raise DeviceFacade.JsonRpcError(e)

            print(COLOR_FAIL + f"Attempted to get text {attempts} times. You may have a slow network or are "
                               f"experiencing another problem." + COLOR_ENDC)
            return ""

        def get_selected(self) -> bool:
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.info["selected"]
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.info["selected"]
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def set_text(self, text):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.set_text(text)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    self.viewV2.set_text(text)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def _double_click_v1(self):
            import uiautomator
            config = self.deviceV1.server.jsonrpc.getConfigurator()
            config['actionAcknowledgmentTimeout'] = 40
            self.deviceV1.server.jsonrpc.setConfigurator(config)
            try:
                self.viewV1.click()
                self.viewV1.click()
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            config['actionAcknowledgmentTimeout'] = 3000
            self.deviceV1.server.jsonrpc.setConfigurator(config)

        def _double_click_v2(self, padding):
            import uiautomator2
            visible_bounds = self.get_bounds()
            horizontal_len = visible_bounds["right"] - visible_bounds["left"]
            vertical_len = visible_bounds["bottom"] - visible_bounds["top"]
            horizintal_padding = int(padding * horizontal_len)
            vertical_padding = int(padding * vertical_len)
            random_x = int(
                uniform(
                    visible_bounds["left"] + horizintal_padding,
                    visible_bounds["right"] - horizintal_padding,
                )
            )
            random_y = int(
                uniform(
                    visible_bounds["top"] + vertical_padding,
                    visible_bounds["bottom"] - vertical_padding,
                )
            )
            time_between_clicks = uniform(0.050, 0.200)
            try:
                self.deviceV2.double_click(random_x, random_y, duration=time_between_clicks)
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)

    @unique
    class Direction(Enum):
        TOP = 0
        BOTTOM = 1
        RIGHT = 2
        LEFT = 3

    @unique
    class Place(Enum):
        # TODO: add more places
        RIGHT = 0
        WHOLE = 1
        CENTER = 2
        BOTTOM = 3
        LEFT = 4

    class JsonRpcError(Exception):
        pass



from insomniac.__version__ import __debug_mode__, __logo__
from insomniac.activation import activation_controller
from insomniac.utils import *


def run(activation_code=""):
    if not __debug_mode__:
        print_timeless(COLOR_OKGREEN + __logo__ + COLOR_ENDC)
        print_version()

    activation_controller.validate(activation_code)
    if not activation_controller.is_activated:
        from insomniac.session import InsomniacSession
        print_timeless("Using insomniac session-manager without extra-features")
        insomniac_session = InsomniacSession()
    else:
        from insomniac.extra_features.session import ExtendedInsomniacSession
        insomniac_session = ExtendedInsomniacSession()

    insomniac_session.run()



from datetime import timedelta
from enum import Enum, unique

from insomniac.database_engine import *
from insomniac.utils import *

FILENAME_INTERACTED_USERS = "interacted_users.json"  # deprecated
FILENAME_SCRAPPED_USERS = "scrapped_users.json"  # deprecated
FILENAME_FILTERED_USERS = "filtered_users.json"  # deprecated
USER_LAST_INTERACTION = "last_interaction"
USER_INTERACTIONS_COUNT = "interactions_count"
USER_FILTERED_AT = "filtered_at"
USER_FOLLOWING_STATUS = "following_status"
USER_SCRAPPING_STATUS = "scrapping_status"

FILENAME_WHITELIST = "whitelist.txt"
FILENAME_BLACKLIST = "blacklist.txt"
FILENAME_TARGETS = "targets.txt"
FILENAME_LOADED_TARGETS = "targets_loaded.txt"
FILENAME_FOLLOWERS = "followers.txt"


STORAGE_ARGS = {
    "reinteract_after": {
        "help": "set a time (in hours) to wait before re-interact with an already interacted profile, disabled by default (won't interact again). "
                "It can be a number (e.g. 48) or a range (e.g. 50-80)",
        "metavar": "150"
    }
}


class Storage:
    database = None
    scrapping_databases = []
    reinteract_after = None
    whitelist = []
    blacklist = []
    account_followers = {}

    def __init__(self, my_username, args):
        self.database = get_database(my_username)

        if args.reinteract_after is not None:
            self.reinteract_after = get_value(args.reinteract_after, "Re-interact after {} hours", 168)

        scrape_for_account = args.__dict__.get('scrape_for_account', None)
        if my_username is None:
            print(COLOR_FAIL + "No username, thus the script won't get access to interacted users and sessions data" +
                  COLOR_ENDC)
            return

        if not os.path.exists(my_username):
            os.makedirs(my_username)

        # Whitelist and Blacklist
        whitelist_path = os.path.join(my_username, FILENAME_WHITELIST)
        if os.path.exists(whitelist_path):
            with open(whitelist_path, encoding="utf-8") as file:
                self.whitelist = [line.rstrip() for line in file]
        blacklist_path = os.path.join(my_username, FILENAME_BLACKLIST)
        if os.path.exists(blacklist_path):
            with open(blacklist_path, encoding="utf-8") as file:
                self.blacklist = [line.rstrip() for line in file]

        # Read targets from targets.txt
        targets_path = os.path.join(my_username, FILENAME_TARGETS)
        if os.path.exists(targets_path):
            with open(targets_path, 'r+', encoding="utf-8") as file_targets:
                targets = [line.rstrip() for line in file_targets]
                # Add targets to the database
                add_targets(self.database, targets, Provider.TARGETS_LIST)
                targets_loaded_path = os.path.join(my_username, FILENAME_LOADED_TARGETS)
                # Add targets to targets_loaded.txt
                with open(targets_loaded_path, 'a+', encoding="utf-8") as file_loaded_targets:
                    for target in targets:
                        file_loaded_targets.write(f"{target}\n")
                # Clear targets.txt
                file_targets.truncate(0)

        # Scraping
        if scrape_for_account is not None:
            for acc in scrape_for_account:
                self.scrapping_databases.append(get_database(acc))

            # TODO: implement 'dump-followers' feature or remove these lines
            # self.followers_path = os.path.join(scrape_for_account, FILENAME_FOLLOWERS)
            # if os.path.exists(self.followers_path):
            #     with open(self.followers_path, encoding="utf-8") as json_file:
            #         self.account_followers = json.load(json_file)

    def check_user_was_interacted(self, username):
        user = get_interacted_user(self.database, username)
        if self.reinteract_after is None:
            return user is not None and user[USER_INTERACTIONS_COUNT] > 0

        return self.check_user_was_interacted_recently(username, hours=self.reinteract_after)

    def check_user_was_interacted_recently(self, username, hours=72):
        user = get_interacted_user(self.database, username)
        if user is None or user[USER_INTERACTIONS_COUNT] == 0:
            return False

        last_interaction = datetime.strptime(user[USER_LAST_INTERACTION], '%Y-%m-%d %H:%M:%S')
        return datetime.now() - last_interaction <= timedelta(hours=hours)

    def check_user_was_scrapped(self, username):
        user = get_scraped_user(self.database, username)
        return user is not None

    def check_user_was_filtered(self, username):
        user = get_filtered_user(self.database, username)
        return user is not None

    def get_following_status(self, username):
        user = get_interacted_user(self.database, username)
        return FollowingStatus.NONE if user is None else FollowingStatus[user[USER_FOLLOWING_STATUS].upper()]

    def add_interacted_user(self,
                            username,
                            last_interaction=datetime.now(),
                            followed=False,
                            unfollowed=False,
                            source=None,
                            interaction_type=None,
                            provider=Provider.UNKNOWN):
        following_status = FollowingStatus.NONE
        if followed:
            following_status = FollowingStatus.FOLLOWED
        if unfollowed:
            following_status = FollowingStatus.UNFOLLOWED
        update_interacted_users(self.database,
                                (username,),
                                (last_interaction,),
                                (following_status,),
                                (source,),
                                (interaction_type,),
                                (provider,))

    def add_scrapped_user(self, username, last_interaction=datetime.now(), success=False):
        scraping_status = ScrappingStatus.SCRAPED if success else ScrappingStatus.NOT_SCRAPED
        update_scraped_users(self.database, (username,), (last_interaction,), (scraping_status,))

    def add_filtered_user(self, username, filtered_at=datetime.now()):
        update_filtered_users(self.database, (username,), (filtered_at,))

    def add_target(self, username, source, interaction_type):
        """
        Add a target to the scrapping_database (it's a database of original account for which we are scrapping).
        """
        for scrapping_database in self.scrapping_databases:
            add_targets(scrapping_database, (username,), Provider.SCRAPING, source, interaction_type)

    def get_target(self):
        """
        Read and remove a target from the targets table.

        :return: a target or None if table is empty.
        """
        return get_target(self.database)

    def save_followers_for_today(self, followers_list, override=False):
        # TODO: implement 'dump-followers' feature or remove this function
        curr_day = str(datetime.date(datetime.now()))
        if curr_day in self.account_followers:
            if not override:
                return

        self.account_followers[curr_day] = followers_list

        if self.followers_path is not None:
            with open(self.followers_path, 'w', encoding="utf-8") as outfile:
                json.dump(self.account_followers, outfile, indent=4, sort_keys=False)

    def is_user_in_whitelist(self, username):
        return username in self.whitelist

    def is_user_in_blacklist(self, username):
        return username in self.blacklist


@unique
class FollowingStatus(Enum):
    NONE = 0
    FOLLOWED = 1
    UNFOLLOWED = 2


@unique
class ScrappingStatus(Enum):
    SCRAPED = 0
    NOT_SCRAPED = 1



from collections import namedtuple
from enum import unique, Enum

LikeAction = namedtuple('LikeAction', 'source user')
FollowAction = namedtuple('FollowAction', 'source user')
UnfollowAction = namedtuple('UnfollowAction', 'user')
StoryWatchAction = namedtuple('StoryWatchAction', 'user')
InteractAction = namedtuple('InteractAction', 'source user succeed')
GetProfileAction = namedtuple('GetProfileAction', 'user')
ScrapeAction = namedtuple('ScrapeAction', 'source user')
RemoveMassFollowerAction = namedtuple('RemoveMassFollowerAction', 'user')


@unique
class BloggerInteractionType(Enum):
    FOLLOWERS = 'followers'
    FOLLOWING = 'following'


@unique
class HashtagInteractionType(Enum):
    TOP_LIKERS = 'top-likers'
    RECENT_LIKERS = 'recent-likers'



from datetime import timedelta

from insomniac.utils import *


def print_full_report(sessions):
    if len(sessions) > 1:
        for index, session in enumerate(sessions):
            finish_time = session.finishTime or datetime.now()
            print_timeless("\n")
            print_timeless(COLOR_REPORT + "SESSION #" + str(index + 1) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Start time: " + str(session.startTime) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Finish time: " + str(finish_time) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Duration: " + str(finish_time - session.startTime) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Total interactions: " + _stringify_interactions(session.totalInteractions)
                           + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Successful interactions: "
                           + _stringify_interactions(session.successfulInteractions) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Total followed: "
                           + _stringify_interactions(session.totalFollowed) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Total likes: " + str(session.totalLikes) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Total unfollowed: " + str(session.totalUnfollowed) + COLOR_ENDC)
            print_timeless(COLOR_REPORT + "Removed mass followers: "
                           + _stringify_removed_mass_followers(session.removedMassFollowers) + COLOR_ENDC)

    print_timeless("\n")
    print_timeless(COLOR_REPORT + "TOTAL" + COLOR_ENDC)

    completed_sessions = [session for session in sessions if session.is_finished()]
    print_timeless(COLOR_REPORT + "Completed sessions: " + str(len(completed_sessions)) + COLOR_ENDC)

    duration = timedelta(0)
    for session in sessions:
        finish_time = session.finishTime or datetime.now()
        duration += finish_time - session.startTime
    print_timeless(COLOR_REPORT + "Total duration: " + str(duration) + COLOR_ENDC)

    total_interactions = {}
    successful_interactions = {}
    total_followed = {}
    total_removed_mass_followers = []
    for session in sessions:
        for source, count in session.totalInteractions.items():
            if total_interactions.get(source) is None:
                total_interactions[source] = count
            else:
                total_interactions[source] += count

        for source, count in session.successfulInteractions.items():
            if successful_interactions.get(source) is None:
                successful_interactions[source] = count
            else:
                successful_interactions[source] += count

        for source, count in session.totalFollowed.items():
            if total_followed.get(source) is None:
                total_followed[source] = count
            else:
                total_followed[source] += count

        for username in session.removedMassFollowers:
            total_removed_mass_followers.append(username)

    print_timeless(COLOR_REPORT + "Total interactions: " + _stringify_interactions(total_interactions) + COLOR_ENDC)
    print_timeless(COLOR_REPORT + "Successful interactions: " + _stringify_interactions(successful_interactions)
                   + COLOR_ENDC)
    print_timeless(COLOR_REPORT + "Total followed : " + _stringify_interactions(total_followed)
                   + COLOR_ENDC)

    total_likes = sum(session.totalLikes for session in sessions)
    print_timeless(COLOR_REPORT + "Total likes: " + str(total_likes) + COLOR_ENDC)

    total_unfollowed = sum(session.totalUnfollowed for session in sessions)
    print_timeless(COLOR_REPORT + "Total unfollowed: " + str(total_unfollowed) + COLOR_ENDC)

    total_story_watches = sum(session.totalStoriesWatched for session in sessions)
    print_timeless(COLOR_REPORT + "Total stories watches: " + str(total_story_watches) + COLOR_ENDC)

    print_timeless(COLOR_REPORT + "Removed mass followers: "
                   + _stringify_removed_mass_followers(total_removed_mass_followers) + COLOR_ENDC)


def print_short_report(source, session_state):
    total_likes = session_state.totalLikes
    total_followed = sum(session_state.totalFollowed.values())
    interactions = session_state.successfulInteractions.get(source, 0)
    total_successful_interactions = sum(session_state.successfulInteractions.values())
    total_story_views = session_state.totalStoriesWatched
    print(COLOR_REPORT + "Session progress: " + str(total_likes) + " likes, " + str(total_followed) + " followed, " +
          str(total_story_views) + " stories watched, " + str(interactions) + " successful " +
          ("interaction" if interactions == 1 else "interactions") +
          " for " + source + "," +
          " " + str(total_successful_interactions) + " successful interactions for the entire session" + COLOR_ENDC)


def print_interaction_types(username, can_like, can_follow, can_watch):
    interaction_types = []
    if can_like:
        interaction_types.append("like")
    if can_follow:
        interaction_types.append("follow")
    if can_watch:
        interaction_types.append("watch stories")
    print(f"@{username} interaction: going to {', '.join(interaction_types)}")


def _stringify_interactions(interactions):
    if len(interactions) == 0:
        return "0"

    result = ""
    total = sum(interactions.values())
    for source, count in interactions.items():
        result += str(count) + " for " + source + ", "
    result = result[:-2]
    return "({0}) {1}".format(total, result)


def _stringify_removed_mass_followers(removed_mass_followers):
    if len(removed_mass_followers) == 0:
        return "none"
    else:
        return "@" + ", @".join(removed_mass_followers)



from insomniac.navigation import switch_to_english
from insomniac.sleeper import sleeper
from insomniac.utils import *
from insomniac.views import TabBarView, ActionBarView


def get_my_profile_info(device):
    try:
        profile_view = TabBarView(device).navigate_to_profile()
        sleeper.random_sleep()

        print("Refreshing your profile status...")
        profile_view.refresh()
        sleeper.random_sleep()

        ActionBarView.update_interaction_rect(device)

        username, followers, following = profile_view.get_profile_info()
    except Exception as e:
        print(COLOR_FAIL + f"Exception: {e}" + COLOR_ENDC)
        save_crash(device, e)
        switch_to_english(device)

        # Try again on the correct language
        profile_view = TabBarView(device).navigate_to_profile()
        sleeper.random_sleep()

        print("Refreshing your profile status...")
        profile_view.refresh()
        sleeper.random_sleep()

        ActionBarView.update_interaction_rect(device)

        username, followers, following = profile_view.get_profile_info()

    report_string = ""
    if username:
        report_string += "Hello, @" + username + "! "
    if followers is not None:
        report_string += "You have " + str(followers) + " followers"
        if following is not None:
            report_string += " and " + str(following) + " followings"
        report_string += " so far."

    if not report_string == "":
        print(report_string)

    return username, followers, following



from functools import partial

from insomniac.action_runners.actions_runners_manager import ActionState
from insomniac.actions_impl import interact_with_user, InteractionStrategy
from insomniac.actions_providers import Provider
from insomniac.actions_types import LikeAction, FollowAction, InteractAction, GetProfileAction, StoryWatchAction, \
    BloggerInteractionType
from insomniac.limits import process_limits
from insomniac.report import print_short_report, print_interaction_types
from insomniac.sleeper import sleeper
from insomniac.softban_indicator import softban_indicator
from insomniac.storage import FollowingStatus
from insomniac.utils import *
from insomniac.views import TabBarView, ProfileView


def extract_blogger_instructions(source):
    split_idx = source.find('-')
    if split_idx == -1:
        print("There is no special interaction-instructions for " + source + ". Working with " + source + " followers.")
        return source, BloggerInteractionType.FOLLOWERS

    selected_instruction = None
    source_profile_name = source[:split_idx]
    interaction_instructions_str = source[split_idx+1:]

    for blogger_instruction in BloggerInteractionType:
        if blogger_instruction.value == interaction_instructions_str:
            selected_instruction = blogger_instruction
            break

    if selected_instruction is None:
        print("Couldn't use interaction-instructions " + interaction_instructions_str +
              ". Working with " + source + " followers.")
        selected_instruction = BloggerInteractionType.FOLLOWERS

    return source_profile_name, selected_instruction


def handle_blogger(device,
                   username,
                   instructions,
                   session_state,
                   likes_count,
                   stories_count,
                   follow_percentage,
                   like_percentage,
                   storage,
                   on_action,
                   is_limit_reached,
                   is_passed_filters,
                   action_status):
    is_myself = username == session_state.my_username
    interaction = partial(interact_with_user,
                          device=device,
                          user_source=username,
                          my_username=session_state.my_username,
                          on_action=on_action)

    search_view = TabBarView(device).navigate_to_search()
    blogger_profile_view = search_view.navigate_to_username(username, on_action)

    if blogger_profile_view is None:
        return

    sleeper.random_sleep()
    is_profile_empty = softban_indicator.detect_empty_profile(device)

    if is_profile_empty:
        return

    followers_following_list_view = None
    if instructions == BloggerInteractionType.FOLLOWERS:
        followers_following_list_view = blogger_profile_view.navigate_to_followers()
    elif instructions == BloggerInteractionType.FOLLOWING:
        followers_following_list_view = blogger_profile_view.navigate_to_following()

    if is_myself:
        followers_following_list_view.scroll_to_bottom()
        followers_following_list_view.scroll_to_top()

    def pre_conditions(follower_name, follower_name_view):
        if storage.is_user_in_blacklist(follower_name):
            print("@" + follower_name + " is in blacklist. Skip.")
            return False
        elif storage.check_user_was_filtered(follower_name):
            print("@" + follower_name + ": already filtered in past. Skip.")
            return False
        elif not is_myself and storage.check_user_was_interacted(follower_name):
            print("@" + follower_name + ": already interacted. Skip.")
            return False
        elif is_myself and storage.check_user_was_interacted_recently(follower_name):
            print("@" + follower_name + ": already interacted in the last week. Skip.")
            return False
        elif is_passed_filters is not None:
            if not is_passed_filters(device, follower_name, reset=True, filters_tags=['BEFORE_PROFILE_CLICK']):
                storage.add_filtered_user(follower_name)
                return False

        return True

    def interact_with_follower(follower_name, follower_name_view):
        is_interact_limit_reached, interact_reached_source_limit, interact_reached_session_limit = \
            is_limit_reached(InteractAction(source=username, user=follower_name, succeed=True), session_state)

        if not process_limits(is_interact_limit_reached, interact_reached_session_limit,
                              interact_reached_source_limit, action_status, "Interaction"):
            return False

        is_get_profile_limit_reached, get_profile_reached_source_limit, get_profile_reached_session_limit = \
            is_limit_reached(GetProfileAction(user=follower_name), session_state)

        if not process_limits(is_get_profile_limit_reached, get_profile_reached_session_limit,
                              get_profile_reached_source_limit, action_status, "Get-Profile"):
            return False

        print("@" + follower_name + ": interact")
        follower_name_view.click()
        on_action(GetProfileAction(user=follower_name))

        sleeper.random_sleep()
        is_profile_empty = softban_indicator.detect_empty_profile(device)

        if is_profile_empty:
            print("Back to followers list")
            device.back()
            return True

        follower_profile_view = ProfileView(device, follower_name == session_state.my_username)

        if is_passed_filters is not None:
            if not is_passed_filters(device, follower_name, reset=False):
                storage.add_filtered_user(follower_name)
                # Continue to next follower
                print("Back to profiles list")
                device.back()
                return True

        is_like_limit_reached, like_reached_source_limit, like_reached_session_limit = \
            is_limit_reached(LikeAction(source=username, user=follower_name), session_state)

        is_follow_limit_reached, follow_reached_source_limit, follow_reached_session_limit = \
            is_limit_reached(FollowAction(source=username, user=follower_name), session_state)

        is_watch_limit_reached, watch_reached_source_limit, watch_reached_session_limit = \
            is_limit_reached(StoryWatchAction(user=follower_name), session_state)

        is_private = follower_profile_view.is_private_account()
        if is_private:
            print("@" + follower_name + ": Private account - images wont be liked.")

        do_have_stories = follower_profile_view.is_story_available()
        if not do_have_stories:
            print("@" + follower_name + ": seems there are no stories to be watched.")

        is_likes_enabled = likes_count != '0'
        is_stories_enabled = stories_count != '0'
        is_follow_enabled = follow_percentage != 0

        likes_value = get_value(likes_count, "Likes count: {}", 2, max_count=12)
        stories_value = get_value(stories_count, "Stories to watch: {}", 1)

        can_like = not is_like_limit_reached and not is_private and likes_value > 0
        can_follow = (not is_follow_limit_reached) and storage.get_following_status(username) == FollowingStatus.NONE and follow_percentage > 0
        can_watch = (not is_watch_limit_reached) and do_have_stories and stories_value > 0
        can_interact = can_like or can_follow or can_watch

        if not can_interact:
            print("@" + follower_name + ": Cant be interacted (due to limits / already followed). Skip.")
            storage.add_interacted_user(follower_name,
                                        followed=False,
                                        source=f"@{username}",
                                        interaction_type=instructions.value,
                                        provider=Provider.INTERACTION)
            on_action(InteractAction(source=username, user=follower_name, succeed=False))
        else:
            print_interaction_types(follower_name, can_like, can_follow, can_watch)
            interaction_strategy = InteractionStrategy(do_like=can_like,
                                                       do_follow=can_follow,
                                                       do_story_watch=can_watch,
                                                       likes_count=likes_value,
                                                       follow_percentage=follow_percentage,
                                                       like_percentage=like_percentage,
                                                       stories_count=stories_value)

            is_liked, is_followed, is_watch = interaction(username=follower_name, interaction_strategy=interaction_strategy)
            if is_liked or is_followed or is_watch:
                storage.add_interacted_user(follower_name,
                                            followed=is_followed,
                                            source=f"@{username}",
                                            interaction_type=instructions.value,
                                            provider=Provider.INTERACTION)
                on_action(InteractAction(source=username, user=follower_name, succeed=True))
                print_short_report(username, session_state)
            else:
                storage.add_interacted_user(follower_name,
                                            followed=False,
                                            source=f"@{username}",
                                            interaction_type=instructions.value,
                                            provider=Provider.INTERACTION)
                on_action(InteractAction(source=username, user=follower_name, succeed=False))

        can_continue = True

        if ((is_like_limit_reached and is_likes_enabled) or not is_likes_enabled) and \
           ((is_follow_limit_reached and is_follow_enabled) or not is_follow_enabled) and \
           ((is_watch_limit_reached and is_stories_enabled) or not is_stories_enabled):
            # If one of the limits reached for source-limit, move to next source
            if (like_reached_source_limit is not None and like_reached_session_limit is None) or \
               (follow_reached_source_limit is not None and follow_reached_session_limit is None):
                can_continue = False
                action_status.set_limit(ActionState.SOURCE_LIMIT_REACHED)

            # If all of the limits reached for session-limit, finish the session
            if ((like_reached_session_limit is not None and is_likes_enabled) or not is_likes_enabled) and \
               ((follow_reached_session_limit is not None and is_follow_enabled) or not is_follow_enabled):
                can_continue = False
                action_status.set_limit(ActionState.SESSION_LIMIT_REACHED)

        print("Back to profiles list")
        device.back()

        return can_continue

    followers_following_list_view.iterate_over_followers(is_myself, interact_with_follower, pre_conditions)



import uuid
from datetime import datetime
from json import JSONEncoder

from insomniac.actions_types import LikeAction, InteractAction, FollowAction, GetProfileAction, ScrapeAction, \
    UnfollowAction, RemoveMassFollowerAction, StoryWatchAction


class SessionState:
    id = None
    args = {}
    app_version = None
    my_username = None
    my_followers_count = None
    my_following_count = None
    totalInteractions = {}
    successfulInteractions = {}
    totalFollowed = {}
    totalLikes = 0
    totalUnfollowed = 0
    totalStoriesWatched = 0
    removedMassFollowers = []
    startTime = None
    finishTime = None

    def __init__(self):
        self.id = str(uuid.uuid4())
        self.args = {}
        self.app_version = None
        self.my_username = None
        self.my_followers_count = None
        self.my_following_count = None
        self.totalInteractions = {}
        self.successfulInteractions = {}
        self.totalFollowed = {}
        self.totalScraped = {}
        self.totalLikes = 0
        self.totalGetProfile = 0
        self.totalUnfollowed = 0
        self.totalStoriesWatched = 0
        self.removedMassFollowers = []
        self.startTime = datetime.now()
        self.finishTime = None

    def add_action(self, action):
        if type(action) == LikeAction:
            self.totalLikes += 1

        if type(action) == GetProfileAction:
            self.totalGetProfile += 1

        if type(action) == StoryWatchAction:
            self.totalStoriesWatched += 1

        if type(action) == InteractAction:
            if self.totalInteractions.get(action.source) is None:
                self.totalInteractions[action.source] = 1
            else:
                self.totalInteractions[action.source] += 1

            if self.successfulInteractions.get(action.source) is None:
                self.successfulInteractions[action.source] = 1 if action.succeed else 0
            else:
                if action.succeed:
                    self.successfulInteractions[action.source] += 1

        if type(action) == FollowAction:
            if self.totalFollowed.get(action.source) is None:
                self.totalFollowed[action.source] = 1
            else:
                self.totalFollowed[action.source] += 1

        if type(action) == ScrapeAction:
            if self.totalScraped.get(action.source) is None:
                self.totalScraped[action.source] = 1
            else:
                self.totalScraped[action.source] += 1

        if type(action) == UnfollowAction:
            self.totalUnfollowed += 1

        if type(action) == RemoveMassFollowerAction:
            self.removedMassFollowers.append(action.user)

    def is_finished(self):
        return self.finishTime is not None



from enum import Enum, unique

from insomniac.actions_types import GetProfileAction
from insomniac.sleeper import sleeper
from insomniac.utils import *

SEARCH_CONTENT_DESC_REGEX = '[Ss]earch and [Ee]xplore'


def navigate(device, tab):
    tab_name = tab.name.lower()
    tab_index = tab.value

    print("Press " + tab_name)
    if tab == Tabs.SEARCH:
        _navigate_to_search(device)
        return

    device.close_keyboard()
    tab_bar = device.find(resourceId=f'{device.app_id}:id/tab_bar', className='android.widget.LinearLayout')
    button = tab_bar.child(index=tab_index)

    # Two clicks to reset tab content
    button.click()
    button.click()


def search_for(device, username=None, hashtag=None, on_action=None):
    navigate(device, Tabs.SEARCH)
    search_edit_text = device.find(resourceId=f'{device.app_id}:id/action_bar_search_edit_text',
                                   className='android.widget.EditText')
    search_edit_text.click()

    if username is not None:
        print("Open user @" + username)
        search_edit_text.set_text(username)
        username_view = device.find(resourceId=f'{device.app_id}:id/row_search_user_username',
                                    className='android.widget.TextView',
                                    text=username)

        sleeper.random_sleep()
        if not username_view.exists():
            print_timeless(COLOR_FAIL + "Cannot find user @" + username + ", abort." + COLOR_ENDC)
            return False

        username_view.click()

        if on_action is not None:
            on_action(GetProfileAction(user=username))

        return True

    if hashtag is not None:
        print("Open hashtag #" + hashtag)
        tab_layout = device.find(resourceId=f'{device.app_id}:id/fixed_tabbar_tabs_container',
                                 className='android.widget.LinearLayout')
        if not tab_layout.exists():
            print(COLOR_FAIL + "Cannot find tabs." + COLOR_ENDC)
            return False
        tab_layout.child(index=2).click()

        search_edit_text.set_text(hashtag)
        hashtag_view = device.find(resourceId=f'{device.app_id}:id/row_hashtag_textview_tag_name',
                                   className='android.widget.TextView',
                                   text=f"#{hashtag}")

        sleeper.random_sleep()
        if not hashtag_view.exists():
            print_timeless(COLOR_FAIL + "Cannot find hashtag #" + hashtag + ", abort." + COLOR_ENDC)
            return False

        hashtag_view.click()
        return True

    return False


def switch_to_english(device):
    print(COLOR_OKGREEN + "Switching to English locale" + COLOR_ENDC)
    navigate(device, Tabs.PROFILE)
    print("Changing language in settings")

    action_bar = device.find(resourceId=f'{device.app_id}:id/action_bar',
                             className='android.widget.LinearLayout')
    # We wanna pick last ImageView in the action bar
    options_view = None
    for options_view in action_bar.child(className='android.widget.ImageView'):
        pass
    if options_view is None or not options_view.exists():
        print(COLOR_FAIL + "No idea how to open menu..." + COLOR_ENDC)
        return
    options_view.click()

    settings_button = device.find(resourceId=f'{device.app_id}:id/menu_settings_row',
                                  className='android.widget.TextView')
    settings_button.click()

    for account_item_index in range(6, 9):
        list_view = device.find(resourceId='android:id/list',
                                className='android.widget.ListView')
        account_item = list_view.child(index=account_item_index)
        account_item.click()

        list_view = device.find(resourceId='android:id/list',
                                className='android.widget.ListView')
        if not list_view.exists():
            print("Opened a wrong tab, going back")
            device.back()
            continue
        language_item = list_view.child(index=4)
        if not language_item.exists():
            print("Opened a wrong tab, going back")
            device.back()
            continue
        language_item.click()

        search_edit_text = device.find(resourceId=f'{device.app_id}:id/search',
                                       className='android.widget.EditText')
        if not search_edit_text.exists():
            print("Opened a wrong tab, going back")
            device.back()
            device.back()
            continue
        search_edit_text.set_text("english")

        list_view = device.find(resourceId=f'{device.app_id}:id/language_locale_list',
                                className='android.widget.ListView')
        english_item = list_view.child(index=0)
        english_item.click()

        break


def _navigate_to_search(device):
    # Search tab is a special case, because on some accounts there is "Reels" tab instead. If so, we have to go to the
    # "Home" tab and press search in the action bar.

    device.close_keyboard()

    tab_bar = device.find(resourceId=f'{device.app_id}:id/tab_bar', className='android.widget.LinearLayout')
    search_in_tab_bar = tab_bar.child(descriptionMatches=SEARCH_CONTENT_DESC_REGEX)
    if search_in_tab_bar.exists():
        # Two clicks to reset tab content
        search_in_tab_bar.click()
        search_in_tab_bar.click()
        return

    print("Didn't find search in the tab bar...")
    navigate(device, Tabs.HOME)
    print("Press search in the action bar")
    action_bar = device.find(resourceId=f'{device.app_id}:id/action_bar', className='android.widget.LinearLayout')
    search_in_action_bar = action_bar.child(descriptionMatches=SEARCH_CONTENT_DESC_REGEX)
    if search_in_action_bar.exists():
        search_in_action_bar.click()
        return

    print(COLOR_FAIL + "Cannot find search tab neither in the tab bar, nor in the action bar. Maybe not English "
                       "language is set?" + COLOR_ENDC)
    save_crash(device)
    switch_to_english(device)
    raise LanguageChangedException()


class LanguageChangedException(Exception):
    pass


@unique
class Tabs(Enum):
    HOME = 0
    SEARCH = 1
    PLUS = 2
    LIKES = 3
    PROFILE = 4



from insomniac.navigation import switch_to_english, LanguageChangedException
from insomniac.utils import *


def parse(device, text):
    multiplier = 1
    text = text.replace(",", "")
    is_dot_in_text = False
    if '.' in text:
        text = text.replace(".", "")
        is_dot_in_text = True
    if "K" in text:
        text = text.replace("K", "")
        multiplier = 1000

        if is_dot_in_text:
            multiplier = 100

    if "M" in text:
        text = text.replace("M", "")
        multiplier = 1000000

        if is_dot_in_text:
            multiplier = 100000
    try:
        count = int(float(text) * multiplier)
    except ValueError as ex:
        print_timeless(COLOR_FAIL + "Cannot parse \"" + text + "\". Probably wrong language, will set English now." +
                       COLOR_ENDC)
        save_crash(device, ex)
        switch_to_english(device)
        raise LanguageChangedException()
    return count



from random import uniform

import speedtest

from insomniac.utils import *

MEGABIT = 1000000
SPEED_GOOD = 25 * MEGABIT
SPEED_BAD = 10 * MEGABIT
SPEED_UGLY = 1 * MEGABIT
SPEED_ZERO = 0

# Dict of sleep ranges by Internet speed
SLEEP_RANGE_BY_SPEED = {
    SPEED_GOOD: (1, 3),
    SPEED_BAD: (2, 5),
    SPEED_UGLY: (4, 8),
    SPEED_ZERO: (7, 12)
}


class Sleeper:
    sleep_range_start = 1.0
    sleep_range_end = 4.0

    def __init__(self):
        pass

    def random_sleep(self):
        delay = uniform(self.sleep_range_start, self.sleep_range_end)
        print(f"Sleep for {delay:.2f} seconds")
        sleep(delay)

    def update_random_sleep_range(self):
        speed = _get_internet_speed()
        if SPEED_ZERO <= speed <= SPEED_UGLY:
            s1 = SPEED_ZERO
            s2 = SPEED_UGLY
        elif SPEED_UGLY <= speed <= SPEED_BAD:
            s1 = SPEED_UGLY
            s2 = SPEED_BAD
        elif SPEED_BAD <= speed <= SPEED_GOOD:
            s1 = SPEED_BAD
            s2 = SPEED_GOOD
        else:
            s1 = SPEED_GOOD
            s2 = SPEED_GOOD

        start1, end1 = SLEEP_RANGE_BY_SPEED[s1]
        start2, end2 = SLEEP_RANGE_BY_SPEED[s2]
        x = (speed - s1) / (s2 - s1) if s2 != s1 else 1  # x is a value between [0, 1]

        self.sleep_range_start = start1 + x * (start2 - start1)
        self.sleep_range_end = end1 + x * (end2 - end1)

        print(f"Sleep range will be from {self.sleep_range_start:.2f} to {self.sleep_range_end:.2f} seconds")


def _get_internet_speed():
    try:
        s = speedtest.Speedtest()
        s.get_best_server()
        s.download(threads=1)
        s.upload(threads=1)
        results_dict = s.results.dict()
    except speedtest.SpeedtestException:
        print(COLOR_FAIL + "Failed to determine Internet speed, supposing it's zero" + COLOR_ENDC)
        return SPEED_ZERO

    download_speed = results_dict['download']
    upload_speed = results_dict['upload']
    print(f"Download speed {(download_speed / MEGABIT):.2f} Mbit/s, upload speed {(upload_speed / MEGABIT):.2f} Mbit/s")
    return (download_speed + upload_speed) / 2


sleeper = Sleeper()



from functools import partial

from insomniac.action_runners.actions_runners_manager import ActionState
from insomniac.actions_impl import interact_with_user, ScrollEndDetector, open_likers, iterate_over_likers, \
    is_private_account, InteractionStrategy, do_have_story
from insomniac.actions_providers import Provider
from insomniac.actions_types import InteractAction, LikeAction, FollowAction, GetProfileAction, StoryWatchAction, \
    HashtagInteractionType
from insomniac.device_facade import DeviceFacade
from insomniac.limits import process_limits
from insomniac.navigation import search_for
from insomniac.report import print_short_report, print_interaction_types
from insomniac.sleeper import sleeper
from insomniac.softban_indicator import softban_indicator
from insomniac.storage import FollowingStatus
from insomniac.utils import *


def extract_hashtag_instructions(source):
    split_idx = source.find('-')
    if split_idx == -1:
        print("There is no special interaction-instructions for " + source + ". Working with " + source + " recent-likers.")
        return source, HashtagInteractionType.RECENT_LIKERS

    selected_instruction = None
    source_profile_name = source[:split_idx]
    interaction_instructions_str = source[split_idx+1:]

    for hashtag_instruction in HashtagInteractionType:
        if hashtag_instruction.value == interaction_instructions_str:
            selected_instruction = hashtag_instruction
            break

    if selected_instruction is None:
        print("Couldn't use interaction-instructions " + interaction_instructions_str +
              ". Working with " + source + " recent-likers.")
        selected_instruction = HashtagInteractionType.RECENT_LIKERS

    return source_profile_name, selected_instruction


def handle_hashtag(device,
                   hashtag,
                   instructions,
                   session_state,
                   likes_count,
                   stories_count,
                   follow_percentage,
                   like_percentage,
                   storage,
                   on_action,
                   is_limit_reached,
                   is_passed_filters,
                   action_status):
    interaction_source = "#{0}".format(hashtag)
    interaction = partial(interact_with_user,
                          device=device,
                          user_source=interaction_source,
                          my_username=session_state.my_username,
                          on_action=on_action)

    def pre_conditions(liker_username, liker_username_view):
        if storage.is_user_in_blacklist(liker_username):
            print("@" + liker_username + " is in blacklist. Skip.")
            return False
        elif storage.check_user_was_filtered(liker_username):
            print("@" + liker_username + ": already filtered in past. Skip.")
            return False
        elif storage.check_user_was_interacted(liker_username):
            print("@" + liker_username + ": already interacted. Skip.")
            return False
        elif is_passed_filters is not None:
            if not is_passed_filters(device, liker_username, reset=True, filters_tags=['BEFORE_PROFILE_CLICK']):
                storage.add_filtered_user(liker_username)
                return False

        return True

    def interact_with_profile(liker_username, liker_username_view):
        is_interact_limit_reached, interact_reached_source_limit, interact_reached_session_limit = \
            is_limit_reached(InteractAction(source=interaction_source, user=liker_username, succeed=True), session_state)

        if not process_limits(is_interact_limit_reached, interact_reached_session_limit,
                              interact_reached_source_limit, action_status, "Interaction"):
            return False

        is_get_profile_limit_reached, get_profile_reached_source_limit, get_profile_reached_session_limit = \
            is_limit_reached(GetProfileAction(user=liker_username), session_state)

        if not process_limits(is_get_profile_limit_reached, get_profile_reached_session_limit,
                              get_profile_reached_source_limit, action_status, "Get-Profile"):
            return False

        print("@" + liker_username + ": interact")
        liker_username_view.click()
        on_action(GetProfileAction(user=liker_username))

        sleeper.random_sleep()
        is_profile_empty = softban_indicator.detect_empty_profile(device)

        if is_profile_empty:
            print("Back to followers list")
            device.back()
            return True

        if is_passed_filters is not None:
            if not is_passed_filters(device, liker_username, reset=False):
                storage.add_filtered_user(liker_username)
                # Continue to next follower
                print("Back to followers list")
                device.back()
                return True

        is_like_limit_reached, like_reached_source_limit, like_reached_session_limit = \
            is_limit_reached(LikeAction(source=interaction_source, user=liker_username), session_state)

        is_follow_limit_reached, follow_reached_source_limit, follow_reached_session_limit = \
            is_limit_reached(FollowAction(source=interaction_source, user=liker_username), session_state)

        is_watch_limit_reached, watch_reached_source_limit, watch_reached_session_limit = \
            is_limit_reached(StoryWatchAction(user=liker_username), session_state)

        is_private = is_private_account(device)
        if is_private:
            print("@" + liker_username + ": Private account - images wont be liked.")

        do_have_stories = do_have_story(device)
        if not do_have_stories:
            print("@" + liker_username + ": seems there are no stories to be watched.")

        is_likes_enabled = likes_count != '0'
        is_stories_enabled = stories_count != '0'
        is_follow_enabled = follow_percentage != 0

        likes_value = get_value(likes_count, "Likes count: {}", 2, max_count=12)
        stories_value = get_value(stories_count, "Stories to watch: {}", 1)

        can_like = not is_like_limit_reached and not is_private and likes_value > 0
        can_follow = (not is_follow_limit_reached) and storage.get_following_status(liker_username) == FollowingStatus.NONE and follow_percentage > 0
        can_watch = (not is_watch_limit_reached) and do_have_stories and stories_value > 0
        can_interact = can_like or can_follow or can_watch

        if not can_interact:
            print("@" + liker_username + ": Cant be interacted (due to limits / already followed). Skip.")
            storage.add_interacted_user(liker_username,
                                        followed=False,
                                        source=f"#{hashtag}",
                                        interaction_type=instructions.value,
                                        provider=Provider.INTERACTION)
            on_action(InteractAction(source=interaction_source, user=liker_username, succeed=False))
        else:
            print_interaction_types(liker_username, can_like, can_follow, can_watch)
            interaction_strategy = InteractionStrategy(do_like=can_like,
                                                       do_follow=can_follow,
                                                       do_story_watch=can_watch,
                                                       likes_count=likes_value,
                                                       follow_percentage=follow_percentage,
                                                       like_percentage=like_percentage,
                                                       stories_count=stories_value)

            is_liked, is_followed, is_watch = interaction(username=liker_username,
                                                          interaction_strategy=interaction_strategy)
            if is_liked or is_followed or is_watch:
                storage.add_interacted_user(liker_username,
                                            followed=is_followed,
                                            source=f"#{hashtag}",
                                            interaction_type=instructions.value,
                                            provider=Provider.INTERACTION)
                on_action(InteractAction(source=interaction_source, user=liker_username, succeed=True))
                print_short_report(interaction_source, session_state)
            else:
                storage.add_interacted_user(liker_username,
                                            followed=False,
                                            source=f"#{hashtag}",
                                            interaction_type=instructions.value,
                                            provider=Provider.INTERACTION)
                on_action(InteractAction(source=interaction_source, user=liker_username, succeed=False))

        can_continue = True

        if ((is_like_limit_reached and is_likes_enabled) or not is_likes_enabled) and \
                ((is_follow_limit_reached and is_follow_enabled) or not is_follow_enabled) and \
                ((is_watch_limit_reached and is_stories_enabled) or not is_stories_enabled):
            # If one of the limits reached for source-limit, move to next source
            if (like_reached_source_limit is not None and like_reached_session_limit is None) or \
                    (follow_reached_source_limit is not None and follow_reached_session_limit is None):
                can_continue = False
                action_status.set_limit(ActionState.SOURCE_LIMIT_REACHED)

            # If all of the limits reached for session-limit, finish the session
            if ((like_reached_session_limit is not None and is_likes_enabled) or not is_likes_enabled) and \
                    ((follow_reached_session_limit is not None and is_follow_enabled) or not is_follow_enabled):
                can_continue = False
                action_status.set_limit(ActionState.SESSION_LIMIT_REACHED)

        print("Back to followers list")
        device.back()

        return can_continue

    extract_hashtag_profiles_and_interact(device,
                                          hashtag,
                                          instructions,
                                          interact_with_profile,
                                          pre_conditions,
                                          on_action)


def extract_hashtag_profiles_and_interact(device,
                                          hashtag,
                                          instructions,
                                          iteration_callback,
                                          iteration_callback_pre_conditions,
                                          on_action):
    print("Interacting with #{0}-{1}".format(hashtag, instructions.value))

    if not search_for(device, hashtag=hashtag, on_action=on_action):
        return

    # Switch to Recent tab
    if instructions == HashtagInteractionType.RECENT_LIKERS:
        print("Switching to Recent tab")
        tab_layout = device.find(resourceId=f'{device.app_id}:id/tab_layout',
                                 className='android.widget.LinearLayout')
        if tab_layout.exists():
            tab_layout.child(index=1).click()
        else:
            print("Can't Find recent tab. Interacting with Popular.")

        sleeper.random_sleep()

    # Open first post
    print("Opening the first post")
    # Index 1 is reserved for hot Reels by this tag
    first_post_index = 2 if instructions == HashtagInteractionType.TOP_LIKERS else 1
    first_post_view = device.find(resourceId=f'{device.app_id}:id/image_button',
                                  className='android.widget.ImageView',
                                  index=first_post_index)
    first_post_view.click()
    sleeper.random_sleep()

    posts_list_view = device.find(resourceId='android:id/list',
                                  className='androidx.recyclerview.widget.RecyclerView')
    posts_end_detector = ScrollEndDetector(repeats_to_end=2)

    def pre_conditions(liker_username, liker_username_view):
        posts_end_detector.notify_username_iterated(liker_username)
        return iteration_callback_pre_conditions(liker_username, liker_username_view)

    while True:
        if not open_likers(device):
            print(COLOR_OKGREEN + "No likes, let's scroll down." + COLOR_ENDC)
            posts_list_view.scroll(DeviceFacade.Direction.BOTTOM)
            continue

        print("List of likers is opened.")
        posts_end_detector.notify_new_page()
        sleeper.random_sleep()

        should_continue_using_source = iterate_over_likers(device, iteration_callback, pre_conditions)

        if not should_continue_using_source:
            break

        if posts_end_detector.is_the_end():
            break
        else:
            posts_list_view.scroll(DeviceFacade.Direction.BOTTOM)



import datetime
from enum import Enum

from insomniac.actions_types import GetProfileAction
from insomniac.device_facade import DeviceFacade
from insomniac.scroll_end_detector import ScrollEndDetector
from insomniac.sleeper import sleeper
from insomniac.utils import *


def case_insensitive_re(str_list):
    if isinstance(str_list, str):
        strings = str_list
    else:
        strings = "|".join(str_list)
    re_str = f"(?i)({strings})"
    return re_str


class TabBarTabs(Enum):
    HOME = 0
    SEARCH = 1
    REELS = 2
    ORDERS = 3
    ACTIVITY = 4
    PROFILE = 5


class SearchTabs(Enum):
    TOP = 0
    ACCOUNTS = 1
    TAGS = 2
    PLACES = 3


class ProfileTabs(Enum):
    POSTS = 0
    IGTV = 1
    REELS = 2
    EFFECTS = 3
    PHOTOS_OF_YOU = 4


class InstagramView:
    def __init__(self, device: DeviceFacade):
        self.device = device

    def is_block_dialog_present(self):
        block_dialog = self.device.find(resourceId=f'{self.device.app_id}:id/dialog_root_view',
                                        className='android.widget.FrameLayout')
        return block_dialog.exists()


class TabBarView(InstagramView):
    HOME_CONTENT_DESC = "Home"
    SEARCH_CONTENT_DESC = "[Ss]earch and [Ee]xplore"
    REELS_CONTENT_DESC = "Reels"
    ORDERS_CONTENT_DESC = "Orders"
    ACTIVITY_CONTENT_DESC = "Activity"
    PROFILE_CONTENT_DESC = "Profile"

    def _get_tab_bar(self):
        self.device.close_keyboard()

        tab_bar = self.device.find(
            resourceIdMatches=case_insensitive_re(f"{self.device.app_id}:id/tab_bar"),
            className="android.widget.LinearLayout",
        )
        return tab_bar

    def navigate_to_home(self):
        self._navigate_to(TabBarTabs.HOME)
        return HomeView(self.device)

    def navigate_to_search(self):
        self._navigate_to(TabBarTabs.SEARCH)
        return SearchView(self.device)

    def navigate_to_reels(self):
        self._navigate_to(TabBarTabs.REELS)

    def navigate_to_orders(self):
        self._navigate_to(TabBarTabs.ORDERS)

    def navigate_to_activity(self):
        self._navigate_to(TabBarTabs.ACTIVITY)

    def navigate_to_profile(self):
        self._navigate_to(TabBarTabs.PROFILE)
        return ProfileView(self.device, is_own_profile=True)

    def _navigate_to(self, tab: TabBarTabs):
        tab_name = tab.name
        print_debug(f"Navigate to {tab_name}")
        button = None
        tab_bar_view = self._get_tab_bar()
        if tab == TabBarTabs.HOME:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.HOME_CONTENT_DESC)
            )
        elif tab == TabBarTabs.SEARCH:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.SEARCH_CONTENT_DESC)
            )
            if not button.exists():
                # Some accounts display the search btn only in Home -> action bar
                print_debug("Didn't find search in the tab bar...")
                home_view = self.navigate_to_home()
                home_view.navigate_to_search()
                return
        elif tab == TabBarTabs.REELS:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.REELS_CONTENT_DESC)
            )
        elif tab == TabBarTabs.ORDERS:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.ORDERS_CONTENT_DESC)
            )
        elif tab == TabBarTabs.ACTIVITY:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.ACTIVITY_CONTENT_DESC)
            )
        elif tab == TabBarTabs.PROFILE:
            button = tab_bar_view.child(
                descriptionMatches=case_insensitive_re(TabBarView.PROFILE_CONTENT_DESC)
            )

        if button.exists():
            # Two clicks to reset tab content
            button.click()
            button.click()

            return

        print(COLOR_FAIL + f"Didn't find tab {tab_name} in the tab bar... "
                           f"Maybe English language is not set!?" + COLOR_ENDC)

        raise LanguageNotEnglishException()


class ActionBarView(InstagramView):
    action_bar_bottom = None
    tab_bar_top = None

    def __init__(self, device: DeviceFacade):
        super().__init__(device)
        self.action_bar = self._get_action_bar()

    def _get_action_bar(self):
        tab_bar = self.device.find(
            resourceIdMatches=case_insensitive_re(f"{self.device.app_id}:id/action_bar_container"),
            className="android.widget.FrameLayout")
        return tab_bar

    @staticmethod
    def update_interaction_rect(device):
        action_bar = device.find(resourceId=f'{device.app_id}:id/action_bar_container',
                                 className='android.widget.FrameLayout')
        if action_bar.exists():
            ActionBarView.action_bar_bottom = action_bar.get_bounds()['bottom']

        device.close_keyboard()

        tab_bar = device.find(resourceId=f'{device.app_id}:id/tab_bar',
                              className='android.widget.LinearLayout')
        if tab_bar.exists():
            ActionBarView.tab_bar_top = tab_bar.get_bounds()['top']

    @staticmethod
    def is_in_interaction_rect(view):
        if ActionBarView.action_bar_bottom is None or ActionBarView.tab_bar_top is None:
            print(COLOR_FAIL + "Interaction rect is not specified." + COLOR_ENDC)
            return False

        view_top = view.get_bounds()['top']
        view_bottom = view.get_bounds()['bottom']
        return ActionBarView.action_bar_bottom <= view_top and view_bottom <= ActionBarView.tab_bar_top


class HomeView(ActionBarView):
    def __init__(self, device: DeviceFacade):
        super().__init__(device)

    def navigate_to_search(self):
        print_debug("Navigate to Search")
        search_btn = self.action_bar.child(
            descriptionMatches=case_insensitive_re(TabBarView.SEARCH_CONTENT_DESC)
        )
        search_btn.click()

        return SearchView(self.device)


class HashTagView(InstagramView):
    def _get_recycler_view(self):
        CLASSNAME = "(androidx.recyclerview.widget.RecyclerView|android.view.View)"

        return self.device.find(classNameMatches=CLASSNAME)

    def _get_first_image_view(self, recycler):
        return recycler.child(
            className="android.widget.ImageView",
            resourceIdMatches=f"{self.device.app_id}:id/image_button",
        )

    def _get_recent_tab(self):
        return self.device.find(
            className="android.widget.TextView",
            text="Recent",
        )


class SearchView(InstagramView):
    def _get_search_edit_text(self):
        return self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/action_bar_search_edit_text"
            ),
            className="android.widget.EditText",
        )

    def _get_username_row(self, username):
        return self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/row_search_user_username"
            ),
            className="android.widget.TextView",
            text=username,
        )

    def _get_hashtag_row(self, hashtag):
        return self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/row_hashtag_textview_tag_name"
            ),
            className="android.widget.TextView",
            text=f"#{hashtag}",
        )

    def _get_tab_text_view(self, tab: SearchTabs):
        tab_layout = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/fixed_tabbar_tabs_container"
            ),
            className="android.widget.LinearLayout",
        )

        tab_text_view = tab_layout.child(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/tab_button_name_text"
            ),
            className="android.widget.TextView",
            textMatches=case_insensitive_re(tab.name),
        )
        return tab_text_view

    def _search_tab_with_text_placeholder(self, tab: SearchTabs):
        tab_layout = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/fixed_tabbar_tabs_container"
            ),
            className="android.widget.LinearLayout",
        )
        search_edit_text = self._get_search_edit_text()

        fixed_text = "Search {}".format(tab.name if tab.name != "TAGS" else "hashtags")
        print_debug(
            "Going to check if the search bar have as placeholder: {}".format(
                fixed_text
            )
        )

        for item in tab_layout.child(
            resourceId=f"{self.device.app_id}:id/tab_button_fallback_icon",
            className="android.widget.ImageView",
        ):
            item.click()
            # random_sleep()

            # Little trick for force-update the ui and placeholder text
            search_edit_text.click()
            self.device.back()

            if self.device.find(
                className="android.widget.TextView",
                textMatches=case_insensitive_re(fixed_text),
            ).exists():
                return item
        return None

    def navigate_to_username(self, username, on_action):
        print_debug("Navigate to profile @" + username)
        search_edit_text = self._get_search_edit_text()
        search_edit_text.click()

        search_edit_text.set_text(username)
        username_view = self._get_username_row(username)

        if not username_view.exists():
            print(COLOR_FAIL + "Cannot find user @" + username + ", abort." + COLOR_ENDC)
            return None

        username_view.click()
        on_action(GetProfileAction(user=username))

        return ProfileView(self.device, is_own_profile=False)

    def navigate_to_hashtag(self, hashtag):
        print(f"Navigate to hashtag {hashtag}")
        search_edit_text = self._get_search_edit_text()
        search_edit_text.click()

        sleeper.random_sleep()
        hashtag_tab = self._get_tab_text_view(SearchTabs.TAGS)
        if not hashtag_tab.exists():
            print_debug(
                "Cannot find tab: Tags. Going to attempt to search for placeholder in all tabs"
            )
            hashtag_tab = self._search_tab_with_text_placeholder(SearchTabs.TAGS)
            if hashtag_tab is None:
                print(COLOR_FAIL + "Cannot find tab: Tags." + COLOR_ENDC)
                save_crash(self.device)
                return None

        hashtag_tab.click()
        sleeper.random_sleep()
        DeviceFacade.back(self.device)
        sleeper.random_sleep()
        # check if that hashtag already exists in the recent search list -> act as human
        hashtag_view_recent = self._get_hashtag_row(hashtag[1:])

        if hashtag_view_recent.exists():
            hashtag_view_recent.click()
            sleeper.random_sleep()
            return HashTagView(self.device)

        print(f"{hashtag} is not in recent searching history..")
        search_edit_text.set_text(hashtag)
        hashtag_view = self._get_hashtag_row(hashtag[1:])

        if not hashtag_view.exists():
            print(COLOR_FAIL + f"Cannot find hashtag {hashtag}, abort." + COLOR_ENDC)
            save_crash(self.device)
            return None

        hashtag_view.click()
        sleeper.random_sleep()

        return HashTagView(self.device)


class PostsViewList(InstagramView):
    def swipe_to_fit_posts(self, first_post):
        """calculate the right swipe amount necessary to swipe to next post in hashtag post view"""
        POST_CONTAINER = f"{self.device.app_id}:id/zoomable_view_container|{self.device.app_id}:id/carousel_media_group"
        displayWidth = self.device.get_info()["displayWidth"]
        if first_post:
            zoomable_view_container = self.device.find(
                resourceIdMatches=POST_CONTAINER
            ).get_bounds()["bottom"]

            print("Scrolled down to see more posts.")
            self.device.swipe_points(
                displayWidth / 2,
                zoomable_view_container - 1,
                displayWidth / 2,
                zoomable_view_container * 2 / 3,
            )
        else:

            gap_view = self.device.find(
                resourceIdMatches=f"{self.device.app_id}:id/gap_view"
            ).get_bounds()["top"]

            self.device.swipe_points(displayWidth / 2, gap_view, displayWidth / 2, 10)
            zoomable_view_container = self.device.find(resourceIdMatches=(POST_CONTAINER))

            zoomable_view_container = self.device.find(
                resourceIdMatches=POST_CONTAINER
            ).get_bounds()["bottom"]

            self.device.swipe_points(
                displayWidth / 2,
                zoomable_view_container - 1,
                displayWidth / 2,
                zoomable_view_container * 2 / 3,
            )
        return

    def check_if_last_post(self, last_description):
        """check if that post has been just interacted"""
        post_description = self.device.find(
            resourceId=f"{self.device.app_id}:id/row_feed_comment_textview_layout"
        )
        if post_description.exists(True):
            new_description = post_description.get_text().upper()
            if new_description == last_description:
                print("This is the last post for this hashtag")
                return True, new_description
            else:
                return False, new_description


class LanguageView(InstagramView):
    def setLanguage(self, language: str):
        print_debug(f"Set language to {language}")
        search_edit_text = self.device.find(
            resourceId=f"{self.device.app_id}:id/search",
            className="android.widget.EditText",
        )
        search_edit_text.set_text(language)

        list_view = self.device.find(
            resourceId=f"{self.device.app_id}:id/language_locale_list",
            className="android.widget.ListView",
        )
        first_item = list_view.child(index=0)
        first_item.click()


class AccountView(InstagramView):
    def navigateToLanguage(self):
        print_debug("Navigate to Language")
        button = self.device.find(
            textMatches=case_insensitive_re("Language"),
            resourceId=f"{self.device.app_id}:id/row_simple_text_textview",
            className="android.widget.TextView",
        )
        button.click()

        return LanguageView(self.device)


class SettingsView(InstagramView):
    def navigateToAccount(self):
        print_debug("Navigate to Account")
        button = self.device.find(
            textMatches=case_insensitive_re("Account"),
            resourceId=f"{self.device.app_id}:id/row_simple_text_textview",
            className="android.widget.TextView",
        )
        button.click()
        return AccountView(self.device)


class OptionsView(InstagramView):
    def navigateToSettings(self):
        print_debug("Navigate to Settings")
        button = self.device.find(
            textMatches=case_insensitive_re("Settings"),
            resourceId=f"{self.device.app_id}:id/menu_settings_row",
            className="android.widget.TextView",
        )
        button.click()
        return SettingsView(self.device)


class OpenedPostView(InstagramView):
    def _getPostLikeButton(self, scroll_to_find=True):
        """Find the like button right bellow a post.
        Note: sometimes the like button from the post above or bellow are
        dumped as well, so we need handle that situation.

        scroll_to_find: if the like button is not found, scroll a bit down
                        to try to find it. Default: True
        """
        BTN_LIKE_RES_ID = f"{self.device.app_id}:id/row_feed_button_like"
        MEDIA_GROUP_RE = case_insensitive_re(
            [
                f"{self.device.app_id}:id/media_group",
                f"{self.device.app_id}:id/carousel_media_group",
            ]
        )
        post_view_area = self.device.find(
            resourceIdMatches=case_insensitive_re("android:id/list")
        )
        if not post_view_area.exists():
            print_debug("Cannot find post recycler view area")
            return None

        post_media_view = self.device.find(
            resourceIdMatches=MEDIA_GROUP_RE,
            className="android.widget.FrameLayout",
        )

        if not post_media_view.exists():
            print_debug("Cannot find post media view area")
            return None

        like_btn_view = post_media_view.down(
            resourceIdMatches=case_insensitive_re(OpenedPostView.BTN_LIKE_RES_ID)
        )

        if like_btn_view.exists():
            # threshold of 30% of the display height
            threshold = int((0.3) * self.device.get_info()["displayHeight"])
            like_btn_top_bound = like_btn_view.get_bounds()["top"]
            is_like_btn_in_the_bottom = like_btn_top_bound > threshold

            if not is_like_btn_in_the_bottom:
                print_debug(
                    f"Like button is to high ({like_btn_top_bound} px). Threshold is {threshold} px"
                )

            post_view_area_bottom_bound = post_view_area.get_bounds()["bottom"]
            is_like_btn_visible = like_btn_top_bound <= post_view_area_bottom_bound
            if not is_like_btn_visible:
                print_debug(
                    f"Like btn out of current clickable area. Like btn top ({like_btn_top_bound}) recycler_view bottom ({post_view_area_bottom_bound})"
                )
        else:
            print_debug("Like button not found bellow the post.")

        if (
            not like_btn_view.exists()
            or not is_like_btn_in_the_bottom
            or not is_like_btn_visible
        ):
            if scroll_to_find:
                print_debug("Try to scroll tiny bit down...")
                # Remember: to scroll down we need to swipe up :)
                self.device.swipe(DeviceFacade.Direction.TOP, scale=0.1)
                like_btn_view = post_media_view.down(
                    resourceIdMatches=case_insensitive_re(
                        OpenedPostView.BTN_LIKE_RES_ID
                    )
                )

            if not scroll_to_find or not like_btn_view.exists():
                print(COLOR_FAIL + "Could not find like button bellow the post" + COLOR_ENDC)
                return None

        return like_btn_view

    def _isPostLiked(self):

        like_btn_view = self._getPostLikeButton()
        if not like_btn_view:
            return False

        return like_btn_view.get_selected()

    def likePost(self, click_btn_like=False):
        MEDIA_GROUP_RE = case_insensitive_re(
            [
                f"{self.device.app_id}:id/media_group",
                f"{self.device.app_id}:id/carousel_media_group",
            ]
        )
        post_media_view = self.device.find(
            resourceIdMatches=MEDIA_GROUP_RE, className="android.widget.FrameLayout"
        )

        if click_btn_like:
            like_btn_view = self._getPostLikeButton()
            if not like_btn_view:
                return False
            like_btn_view.click()
        else:

            if post_media_view.exists():
                post_media_view.double_click()
            else:
                print(COLOR_FAIL + "Could not find post area to double click" + COLOR_ENDC)
                return False

        sleeper.random_sleep()

        return self._isPostLiked()

    def open_likers(self):
        while True:
            likes_view = self.device.find(
                resourceId=f"{self.device.app_id}:id/row_feed_textview_likes",
                className="android.widget.TextView",
            )
            if likes_view.exists(True):
                if likes_view.get_text()[-6:].upper() == "OTHERS":
                    print("Opening post likers")
                    sleeper.random_sleep()
                    likes_view.click(likes_view.Location.RIGHT)
                    return True
                else:
                    print("This post has only 1 liker, skip")
                    return False
            else:
                return False

    def _getListViewLikers(self):
        return self.device.find(
            resourceId="android:id/list", className="android.widget.ListView"
        )

    def _getUserCountainer(self):
        return self.device.find(
            resourceId=f"{self.device.app_id}:id/row_user_container_base",
            className="android.widget.LinearLayout",
        )

    def _getUserName(self, countainer):
        return countainer.child(
            resourceId=f"{self.device.app_id}:id/row_user_primary_name",
            className="android.widget.TextView",
        )


class PostsGridView(InstagramView):
    def scrollDown(self):
        coordinator_layout = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/coordinator_root_layout"
            )
        )
        if coordinator_layout.exists():
            coordinator_layout.scroll(DeviceFacade.Direction.BOTTOM)
            return True

        return False

    def navigateToPost(self, row, col):
        post_list_view = self.device.find(
            resourceIdMatches=case_insensitive_re("android:id/list")
        )
        OFFSET = 1  # row with post starts from index 1
        row_view = post_list_view.child(index=row + OFFSET)
        if not row_view.exists():
            return None
        post_view = row_view.child(index=col)
        if not post_view.exists():
            return None
        post_view.click()

        return OpenedPostView(self.device)


class ProfileView(ActionBarView):
    def __init__(self, device: DeviceFacade, is_own_profile=False):
        super().__init__(device)
        self.is_own_profile = is_own_profile

    def refresh(self):
        re_case_insensitive = case_insensitive_re(
            [
                f"{self.device.app_id}:id/coordinator_root_layout",
            ]
        )
        coordinator_layout = self.device.find(resourceIdMatches=re_case_insensitive)
        if coordinator_layout.exists():
            coordinator_layout.scroll(DeviceFacade.Direction.TOP)

    def navigate_to_options(self):
        print_debug("Navigate to Options")
        button = self.action_bar.child(
            descriptionMatches=case_insensitive_re("Options")
        )
        button.click()

        return OptionsView(self.device)

    def _get_action_bar_title_btn(self):
        re_case_insensitive = case_insensitive_re(
            [
                f"{self.device.app_id}:id/title_view",
                f"{self.device.app_id}:id/action_bar_title",
                f"{self.device.app_id}:id/action_bar_large_title",
                f"{self.device.app_id}:id/action_bar_textview_title",
            ]
        )
        return self.action_bar.child(
            resourceIdMatches=re_case_insensitive, className="android.widget.TextView"
        )

    def get_username(self, error=True):
        title_view = self._get_action_bar_title_btn()
        if title_view.exists():
            return title_view.get_text()
        if error:
            print(COLOR_FAIL + "Cannot get username" + COLOR_ENDC)
        return None

    def _parse_counter(self, text):
        multiplier = 1
        text = text.replace(",", "")
        is_dot_in_text = False
        if '.' in text:
            text = text.replace(".", "")
            is_dot_in_text = True
        if "K" in text:
            text = text.replace("K", "")
            multiplier = 1000

            if is_dot_in_text:
                multiplier = 100

        if "M" in text:
            text = text.replace("M", "")
            multiplier = 1000000

            if is_dot_in_text:
                multiplier = 100000
        try:
            count = int(float(text) * multiplier)
        except ValueError:
            print(COLOR_FAIL + f"Cannot parse {text}. Probably wrong language ?!" + COLOR_ENDC)
            raise LanguageNotEnglishException()
        return count

    def _get_followers_text_view(self):
        followers_text_view = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/row_profile_header_textview_followers_count"
            ),
            className="android.widget.TextView",
        )
        return followers_text_view

    def get_followers_count(self, should_parse=True):
        followers = None
        followers_text_view = self._get_followers_text_view()
        if followers_text_view.exists():
            followers_text = followers_text_view.get_text()
            if followers_text:
                if should_parse:
                    followers = self._parse_counter(followers_text)
                else:
                    followers = followers_text
            else:
                print(COLOR_FAIL + "Cannot get followers count text" + COLOR_ENDC)
        else:
            print(COLOR_FAIL + "Cannot find followers count view" + COLOR_ENDC)

        return followers

    def _get_following_text_view(self):
        following_text_view = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/row_profile_header_textview_following_count"
            ),
            className="android.widget.TextView",
        )
        return following_text_view

    def get_following_count(self):
        following = None
        following_text_view = self._get_following_text_view()
        if following_text_view.exists():
            following_text = following_text_view.get_text()
            if following_text:
                following = self._parse_counter(following_text)
            else:
                print(COLOR_FAIL + "Cannot get following count text" + COLOR_ENDC)
        else:
            print(COLOR_FAIL + "Cannot find following count view" + COLOR_ENDC)

        return following

    def get_posts_count(self):
        post_count_view = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/row_profile_header_textview_post_count"
            ),
            className="android.widget.TextView",
        )
        if post_count_view.exists():
            count = post_count_view.get_text()
            if count != None:
                return self._parse_counter(count)
            else:
                print(COLOR_FAIL + "Cannot get posts count text" + COLOR_ENDC)
                return 0
        else:
            print(COLOR_FAIL + "Cannot get posts count text" + COLOR_ENDC)
            return 0

    def count_photo_in_view(self):
        """return rows filled and the number of post in the last row"""
        RECYCLER_VIEW = "androidx.recyclerview.widget.RecyclerView"
        grid_post = self.device.find(
            className=RECYCLER_VIEW, resourceIdMatches="android:id/list"
        )
        if grid_post.exists():  # max 4 rows supported
            for i in range(2, 5):
                lin_layout = grid_post.child(
                    index=i, className="android.widget.LinearLayout"
                )
                if i == 4 or not lin_layout.exists(True):
                    last_index = i - 1
                    last_lin_layout = grid_post.child(index=last_index)
                    for n in range(1, 4):
                        if n == 3 or not last_lin_layout.child(index=n).exists(True):
                            if n == 3:
                                return last_index, 0
                            else:
                                return last_index - 1, n
        else:
            return 0, 0

    def get_profile_info(self):

        username = self.get_username()
        followers = self.get_followers_count()
        following = self.get_following_count()

        return username, followers, following

    def get_profile_biography(self):
        biography = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/profile_header_bio_text"
            ),
            className="android.widget.TextView",
        )
        if biography.exists():
            biography_text = biography.get_text()
            # If the biography is very long, blabla text and end with "...more" click the bottom of the text and get the new text
            is_long_bio = re.compile(
                r"{0}$".format("… more"), flags=re.IGNORECASE
            ).search(biography_text)
            if is_long_bio is not None:
                biography.click(biography.Location.BOTTOM)
                return biography.get_text()
            return biography_text
        return ""

    def get_full_name(self):
        full_name_view = self.device.find(
            resourceIdMatches=case_insensitive_re(
                f"{self.device.app_id}:id/profile_header_full_name"
            ),
            className="android.widget.TextView",
        )
        if full_name_view.exists():
            fullname_text = full_name_view.get_text()
            if fullname_text is not None:
                return fullname_text
        return ""

    def is_private_account(self):
        private_profile_view = self.device.find(
            resourceIdMatches=case_insensitive_re(
                [
                    f"{self.device.app_id}:id/private_profile_empty_state",
                    f"{self.device.app_id}:id/row_profile_header_empty_profile_notice_title",
                ]
            )
        )
        return private_profile_view.exists(quick=True)

    def is_story_available(self):
        return self.device.find(
            resourceId=f"{self.device.app_id}:id/reel_ring",
            className="android.view.View",
        ).exists(quick=True)

    def profile_image(self):
        return self.device.find(
            resourceId=f"{self.device.app_id}:id/row_profile_header_imageview",
            className="android.widget.ImageView",
        )

    def navigate_to_followers(self):
        print_debug("Navigate to Followers")
        FOLLOWERS_BUTTON_ID_REGEX = case_insensitive_re(
            [
                f"{self.device.app_id}:id/row_profile_header_followers_container",
                f"{self.device.app_id}:id/row_profile_header_container_followers",
            ]
        )
        followers_button = self.device.find(resourceIdMatches=FOLLOWERS_BUTTON_ID_REGEX)
        followers_button.click()

        return FollowersFollowingListView(self.device)

    def navigate_to_following(self):
        print_debug("Navigate to Followers")
        FOLLOWING_BUTTON_ID_REGEX = case_insensitive_re(
            [
                f"{self.device.app_id}:id/row_profile_header_following_container",
                f"{self.device.app_id}:id/row_profile_header_container_following",
            ]
        )
        following_button = self.device.find(resourceIdMatches=FOLLOWING_BUTTON_ID_REGEX)
        following_button.click()

        return FollowersFollowingListView(self.device)

    def swipe_to_fit_posts(self):
        """calculate the right swipe amount necessary to see 12 photos"""
        displayWidth = self.device.get_info()["displayWidth"]
        element_to_swipe_over = self.device.find(
            resourceIdMatches=f"{self.device.app_id}:id/profile_tabs_container"
        ).get_bounds()["top"]
        bar_countainer = self.device.find(
            resourceIdMatches=f"{self.device.app_id}:id/action_bar_container"
        ).get_bounds()["bottom"]

        print("Scrolled down to see more posts.")
        self.device.swipe_points(
            displayWidth / 2, element_to_swipe_over, displayWidth / 2, bar_countainer
        )
        return

    def navigate_to_posts_tab(self):
        self._navigate_to_tab(ProfileTabs.POSTS)
        return PostsGridView(self.device)

    def navigate_to_igtv_tab(self):
        self._navigate_to_tab(ProfileTabs.IGTV)
        raise Exception("Not implemented")

    def navigate_to_reels_tab(self):
        self._navigate_to_tab(ProfileTabs.REELS)
        raise Exception("Not implemented")

    def navigate_to_effects_tab(self):
        self._navigate_to_tab(ProfileTabs.EFFECTS)
        raise Exception("Not implemented")

    def navigate_to_photos_of_you_tab(self):
        self._navigate_to_tab(ProfileTabs.PHOTOS_OF_YOU)
        raise Exception("Not implemented")

    def _navigate_to_tab(self, tab: ProfileTabs):
        TABS_RES_ID = f"{self.device.app_id}:id/profile_tab_layout"
        TABS_CLASS_NAME = "android.widget.HorizontalScrollView"
        tabs_view = self.device.find(
            resourceIdMatches=case_insensitive_re(TABS_RES_ID),
            className=TABS_CLASS_NAME,
        )

        TAB_RES_ID = f"{self.device.app_id}:id/profile_tab_icon_view"
        TAB_CLASS_NAME = "android.widget.ImageView"
        description = ""
        if tab == ProfileTabs.POSTS:
            description = "Grid View"
        elif tab == ProfileTabs.IGTV:
            description = "IGTV"
        elif tab == ProfileTabs.REELS:
            description = "Reels"
        elif tab == ProfileTabs.EFFECTS:
            description = "Effects"
        elif tab == ProfileTabs.PHOTOS_OF_YOU:
            description = "Photos of You"

        button = tabs_view.child(
            descriptionMatches=case_insensitive_re(description),
            resourceIdMatches=case_insensitive_re(TAB_RES_ID),
            className=TAB_CLASS_NAME,
        )

        attempts = 0
        while not button.exists():
            attempts += 1
            self.device.swipe(DeviceFacade.Direction.TOP, scale=0.1)
            if attempts > 2:
                print(COLOR_FAIL + f"Cannot navigate to tab '{description}'" + COLOR_ENDC)
                save_crash(self.device)
                return

        button.click()

    def _get_recycler_view(self):
        CLASSNAME = "(androidx.recyclerview.widget.RecyclerView|android.view.View)"

        return self.device.find(classNameMatches=CLASSNAME)


class FollowersFollowingListView(InstagramView):
    def scroll_to_bottom(self):
        print("Scroll to the bottom of the list")

        def is_end_reached():
            see_all_button = self.device.find(resourceId=f'{self.device.app_id}:id/see_all_button',
                                              className='android.widget.TextView')
            return see_all_button.exists()

        list_view = self.device.find(resourceId='android:id/list',
                                     className='android.widget.ListView')
        while not is_end_reached():
            list_view.swipe(DeviceFacade.Direction.BOTTOM)

    def scroll_to_top(self):
        print("Scroll to top of the list")

        def is_at_least_one_follower():
            follower = self.device.find(resourceId=f'{self.device.app_id}:id/follow_list_container',
                                        className='android.widget.LinearLayout')
            return follower.exists()

        list_view = self.device.find(resourceId='android:id/list',
                                     className='android.widget.ListView')

        while not is_at_least_one_follower():
            list_view.scroll(DeviceFacade.Direction.TOP)

    def is_list_empty(self):
        # Looking for any profile in the list, just to make sure its loaded with profiles
        any_username_view = self.device.find(resourceId=f'{self.device.app_id}:id/follow_list_username',
                                             className='android.widget.TextView')
        is_list_empty_from_profiles = not any_username_view.exists()
        return is_list_empty_from_profiles

    def iterate_over_followers(self, is_myself, iteration_callback,
                               iteration_callback_pre_conditions, iterate_without_sleep=False):
        FOLLOW_LIST_CONTAINER_ID_REGEX = case_insensitive_re([f"{self.device.app_id}:id/follow_list_container"])
        ROW_SEARCH_ID_REGEX = case_insensitive_re([f"{self.device.app_id}:id/row_search_edit_text"])
        LOAD_MORE_BUTTON_ID_REGEX = case_insensitive_re([f"{self.device.app_id}:id/row_load_more_button"])

        # Wait until list is rendered
        self.device.find(resourceIdMatches=FOLLOW_LIST_CONTAINER_ID_REGEX,
                         className='android.widget.LinearLayout').wait()

        def scrolled_to_top():
            row_search = self.device.find(resourceIdMatches=ROW_SEARCH_ID_REGEX,
                                          className='android.widget.EditText')
            return row_search.exists()

        prev_screen_iterated_followers = []
        scroll_end_detector = ScrollEndDetector()
        while True:
            print("Iterate over visible followers")
            if not iterate_without_sleep:
                sleeper.random_sleep()

            screen_iterated_followers = []
            screen_skipped_followers_count = 0
            scroll_end_detector.notify_new_page()

            try:
                for item in self.device.find(resourceIdMatches=FOLLOW_LIST_CONTAINER_ID_REGEX,
                                             className='android.widget.LinearLayout'):
                    user_info_view = item.child(index=1)
                    user_name_view = user_info_view.child(index=0).child()
                    if not user_name_view.exists(quick=True):
                        print(COLOR_OKGREEN + "Next item not found: probably reached end of the screen." + COLOR_ENDC)
                        break

                    try:
                        username = user_name_view.get_text()
                    except DeviceFacade.JsonRpcError:
                        print(COLOR_OKGREEN + "Next item was found, but its text is half-cutted / invisible, "
                                              "moving on to next row." + COLOR_ENDC)
                        continue

                    screen_iterated_followers.append(username)
                    scroll_end_detector.notify_username_iterated(username)

                    if not iteration_callback_pre_conditions(username, user_name_view):
                        screen_skipped_followers_count += 1
                        continue

                    to_continue = iteration_callback(username, user_name_view)
                    if not to_continue:
                        print(COLOR_OKBLUE + "Stopping followers iteration" + COLOR_ENDC)
                        return

            except IndexError:
                print(COLOR_FAIL + "Cannot get next item: probably reached end of the screen." + COLOR_ENDC)

            if is_myself and scrolled_to_top():
                print(COLOR_OKGREEN + "Scrolled to top, finish." + COLOR_ENDC)
                return
            elif len(screen_iterated_followers) > 0:
                load_more_button = self.device.find(resourceIdMatches=LOAD_MORE_BUTTON_ID_REGEX)
                load_more_button_exists = load_more_button.exists(quick=True)

                if scroll_end_detector.is_the_end():
                    return

                need_swipe = screen_skipped_followers_count == len(screen_iterated_followers)
                list_view = self.device.find(resourceId='android:id/list',
                                             className='android.widget.ListView')
                if not list_view.exists():
                    print(COLOR_FAIL + "Cannot find the list of followers. Trying to press back again." + COLOR_ENDC)
                    self.device.back()
                    list_view = self.device.find(resourceId='android:id/list',
                                                 className='android.widget.ListView')

                if is_myself:
                    print(COLOR_OKGREEN + "Need to scroll now" + COLOR_ENDC)
                    list_view.scroll(DeviceFacade.Direction.TOP)
                else:
                    pressed_retry = False
                    if load_more_button_exists:
                        retry_button = load_more_button.child(className='android.widget.ImageView')
                        if retry_button.exists():
                            print("Press \"Load\" button")
                            retry_button.click()
                            sleeper.random_sleep()
                            pressed_retry = True

                    if need_swipe and not pressed_retry:
                        print(COLOR_OKGREEN + "All followers skipped, let's do a swipe" + COLOR_ENDC)
                        list_view.swipe(DeviceFacade.Direction.BOTTOM)
                    else:
                        print(COLOR_OKGREEN + "Need to scroll now" + COLOR_ENDC)
                        list_view.scroll(DeviceFacade.Direction.BOTTOM)

                prev_screen_iterated_followers.clear()
                prev_screen_iterated_followers += screen_iterated_followers
            else:
                print(COLOR_OKGREEN + "No followers were iterated, finish." + COLOR_ENDC)
                return


class CurrentStoryView(InstagramView):
    def getStoryFrame(self):
        return self.device.find(
            resourceId=f"{self.device.app_id}:id/reel_viewer_image_view",
            className="android.widget.FrameLayout",
        )

    def getUsername(self):
        reel_viewer_title = self.device.find(
            resourceId=f"{self.device.app_id}:id/reel_viewer_title",
            className="android.widget.TextView",
        )
        return "" if not reel_viewer_title.exists() else reel_viewer_title.get_text()

    def getTimestamp(self):
        reel_viewer_timestamp = self.device.find(
            resourceId=f"{self.device.app_id}:id/reel_viewer_timestamp",
            className="android.widget.TextView",
        )
        if reel_viewer_timestamp.exists():
            timestamp = reel_viewer_timestamp.get_text().strip()
            value = int(re.sub("[^0-9]", "", timestamp))
            if timestamp[-1] == "s":
                return datetime.timestamp(
                    datetime.datetime.now() - datetime.timedelta(seconds=value)
                )
            elif timestamp[-1] == "m":
                return datetime.timestamp(
                    datetime.datetime.now() - datetime.timedelta(minutes=value)
                )
            elif timestamp[-1] == "h":
                return datetime.timestamp(
                    datetime.datetime.now() - datetime.timedelta(hours=value)
                )
            else:
                return datetime.timestamp(
                    datetime.datetime.now() - datetime.timedelta(days=value)
                )
        return None


class LanguageNotEnglishException(Exception):
    pass



import random
import sys

import colorama

import insomniac.__version__
from insomniac.__version__ import __debug_mode__
from insomniac.action_get_my_profile_info import get_my_profile_info
from insomniac.action_runners.actions_runners_manager import ActionRunnersManager
from insomniac.device import DeviceWrapper
from insomniac.limits import LimitsManager
from insomniac.migration import migrate_from_json_to_sql
from insomniac.params import parse_arguments, refresh_args_by_conf_file
from insomniac.report import print_full_report
from insomniac.session_state import SessionState
from insomniac.sessions import Sessions
from insomniac.sleeper import sleeper
from insomniac.softban_indicator import ActionBlockedError
from insomniac.storage import STORAGE_ARGS, Storage
from insomniac.utils import *

sessions = Sessions()


class InsomniacSession(object):
    SESSION_ARGS = {
        "repeat": {
            "help": 'repeat the same session again after N minutes after completion, disabled by default. '
                    'It can be a number of minutes (e.g. 180) or a range (e.g. 120-180)',
            "metavar": '120-180'
        },
        "device": {
            "help": 'device identifier. Should be used only when multiple devices are connected at once',
            "metavar": '2443de990e017ece'
        },
        "wait_for_device": {
            'help': 'keep waiting for ADB-device to be ready for connection (if no device-id is provided using '
                    '--device flag, will wait for any available device)',
            'action': 'store_true',
            "default": False
        },
        "no_speed_check": {
            'help': 'skip internet speed check at start',
            'action': 'store_true'
        },
        "old": {
            'help': 'add this flag to use an old version of uiautomator. Use it only if you experience '
                    'problems with the default version',
            'action': 'store_true'
        },
        "app_id": {
            "help": 'apk package identifier. Should be used only if you are using cloned-app. '
                    'Using \'com.instagram.android\' by default',
            "metavar": 'com.instagram.android',
            "default": 'com.instagram.android'
        },
        "dont_indicate_softban": {
            "help": "by default, Insomniac tried to indicate if there is a softban on your acoount. set this flag in "
                    "order to ignore those soft-ban indicators",
            'action': 'store_true',
            "default": False
        },
        "debug": {
            'help': 'add this flag to insomniac in debug mode (more verbose logs)',
            'action': 'store_true'
        }
    }

    repeat = None
    device = None
    old = None

    def __init__(self):
        random.seed()
        colorama.init()

        self.sessions = sessions
        self.storage = None
        self.session_state = None
        self.actions_mgr = ActionRunnersManager()
        self.limits_mgr = LimitsManager()

    def get_session_args(self):
        all_args = {}
        all_args.update(self.SESSION_ARGS)
        all_args.update(STORAGE_ARGS)
        all_args.update(self.actions_mgr.get_actions_args())
        all_args.update(self.limits_mgr.get_limits_args())

        return all_args

    def set_session_args(self, args):
        if args.repeat is not None:
            self.repeat = get_value(args.repeat, "Sleep time (min) before repeat: {}", 180)

        if args.debug is not None:
            insomniac.__version__.__debug_mode__ = True

        if args.dont_indicate_softban:
            insomniac.softban_indicator.should_indicate_softban = False

    def parse_args_and_get_device_wrapper(self):
        ok, args = parse_arguments(self.get_session_args())
        if not ok:
            return None, None, None

        device_wrapper = DeviceWrapper(args.device, args.old, args.wait_for_device, args.app_id)
        device = device_wrapper.get()
        if device is None:
            return None, None, None

        app_version = get_instagram_version(args.device, args.app_id)

        print("Instagram version: " + app_version)

        return args, device_wrapper, app_version

    def start_session(self, args, device_wrapper, app_version):
        self.session_state = SessionState()
        self.session_state.args = args.__dict__
        self.session_state.app_version = app_version
        self.sessions.append(self.session_state)

        device_wrapper.get().wake_up()

        print_timeless(COLOR_REPORT + "\n-------- START: " + str(self.session_state.startTime) + " --------" + COLOR_ENDC)

        open_instagram(args.device, args.app_id)
        sleeper.random_sleep()
        self.session_state.my_username, \
            self.session_state.my_followers_count, \
            self.session_state.my_following_count = get_my_profile_info(device_wrapper.get())

        return self.session_state

    def end_session(self, device_wrapper):
        close_instagram(device_wrapper.device_id, device_wrapper.app_id)
        print_copyright()
        self.session_state.finishTime = datetime.now()
        print_timeless(COLOR_REPORT + "-------- FINISH: " + str(self.session_state.finishTime) + " --------" + COLOR_ENDC)

        print_full_report(self.sessions)
        print_timeless("")
        self.sessions.persist(self.session_state.my_username)

    def repeat_session(self, args):
        print("Sleep for {} minutes".format(self.repeat))
        try:
            sleep(60 * self.repeat)
            refresh_args_by_conf_file(args)
        except KeyboardInterrupt:
            print_full_report(self.sessions)
            self.sessions.persist(self.session_state.my_username)
            sys.exit(0)

    def on_action_callback(self, action):
        self.session_state.add_action(action)
        self.limits_mgr.update_state(action)

    def run(self):
        args, device_wrapper, app_version = self.parse_args_and_get_device_wrapper()
        if args is None or device_wrapper is None:
            return

        if not args.no_speed_check:
            print("Checking your Internet speed to adjust the script speed, please wait for a minute...")
            print("(use " + COLOR_BOLD + "--no-speed-check" + COLOR_ENDC + " to skip this check)")
            sleeper.update_random_sleep_range()

        while True:
            self.set_session_args(args)

            action_runner = self.actions_mgr.select_action_runner(args)

            if action_runner is None:
                return

            action_runner.set_params(args)
            self.limits_mgr.set_limits(args)

            try:
                self.start_session(args, device_wrapper, app_version)
                migrate_from_json_to_sql(self.session_state.my_username)
                self.storage = Storage(self.session_state.my_username, args)

                action_runner.run(device_wrapper,
                                  self.storage,
                                  self.session_state,
                                  self.on_action_callback,
                                  self.limits_mgr.is_limit_reached_for_action)
            except KeyboardInterrupt:
                self.end_session(device_wrapper)
                return
            except ActionBlockedError as ex:
                print_timeless("")
                print_timeless(COLOR_FAIL + str(ex) + COLOR_ENDC)
                self.end_session(device_wrapper)
                return
            except Exception as ex:
                if __debug_mode__:
                    raise ex
                else:
                    print_timeless(COLOR_FAIL + f"\nCaught an exception:\n{ex}" + COLOR_ENDC)
                    save_crash(device_wrapper.get(), ex)
            
            self.end_session(device_wrapper)
            if self.repeat is not None:
                self.repeat_session(args)
            else:
                break



from insomniac.session_state import SessionState
from insomniac.sessions import FILENAME_SESSIONS, Sessions
from insomniac.storage import *
from insomniac.utils import *


def migrate_from_json_to_sql(my_username):
    """
    Migration from JSON storage (v3.4.2) to SQL storage (v3.5.0).
    """
    if check_database_exists(my_username):
        return

    database = get_database(my_username)

    interacted_users_path = os.path.join(my_username, FILENAME_INTERACTED_USERS)
    if os.path.exists(interacted_users_path):
        try:
            print(f"[Migration] Loading data from {FILENAME_INTERACTED_USERS}...")
            with open(interacted_users_path, encoding="utf-8") as json_file:
                interacted_users = json.load(json_file)
                usernames = []
                last_interactions = []
                following_statuses = []
                sources = []
                interaction_types = []
                providers = []
                for username in interacted_users.keys():
                    usernames.append(username)
                    user = interacted_users[username]
                    last_interactions.append(datetime.strptime(user[USER_LAST_INTERACTION], '%Y-%m-%d %H:%M:%S.%f'))
                    following_statuses.append(FollowingStatus[user[USER_FOLLOWING_STATUS].upper()])
                    sources.append(None)
                    interaction_types.append(None)
                    providers.append(Provider.UNKNOWN)

                update_interacted_users(database, usernames, last_interactions, following_statuses, sources, interaction_types, providers)
                print(COLOR_BOLD + f"[Migration] Done! You can now delete {FILENAME_INTERACTED_USERS}" + COLOR_ENDC)
        except ValueError as e:
            print(COLOR_FAIL + f"[Migration] Cannot load {FILENAME_INTERACTED_USERS}: {e}" + COLOR_ENDC)

    scrapped_users_path = os.path.join(my_username, FILENAME_SCRAPPED_USERS)
    if os.path.exists(scrapped_users_path):
        try:
            print(f"[Migration] Loading data from {FILENAME_SCRAPPED_USERS}...")
            with open(scrapped_users_path, encoding="utf-8") as json_file:
                scrapped_users = json.load(json_file)
                usernames = []
                last_interactions = []
                scraping_statuses = []
                for username in scrapped_users.keys():
                    usernames.append(username)
                    user = scrapped_users[username]
                    last_interactions.append(datetime.strptime(user[USER_LAST_INTERACTION], '%Y-%m-%d %H:%M:%S.%f'))
                    scraping_statuses.append(ScrappingStatus[user[USER_SCRAPPING_STATUS].upper()])

                update_scraped_users(database, usernames, last_interactions, scraping_statuses)
                print(COLOR_BOLD + f"[Migration] Done! You can now delete {FILENAME_SCRAPPED_USERS}" + COLOR_ENDC)
        except ValueError as e:
            print(COLOR_FAIL + f"[Migration] Cannot load {FILENAME_SCRAPPED_USERS}: {e}" + COLOR_ENDC)

    filtered_users_path = os.path.join(my_username, FILENAME_FILTERED_USERS)
    if os.path.exists(filtered_users_path):
        try:
            print(f"[Migration] Loading data from {FILENAME_FILTERED_USERS}...")
            with open(filtered_users_path, encoding="utf-8") as json_file:
                filtered_users = json.load(json_file)
                usernames = []
                filtered_at_list = []
                for username in filtered_users.keys():
                    usernames.append(username)
                    user = filtered_users[username]
                    filtered_at_list.append(datetime.strptime(user[USER_FILTERED_AT], '%Y-%m-%d %H:%M:%S.%f'))

                update_filtered_users(database, usernames, filtered_at_list)
                print(COLOR_BOLD + f"[Migration] Done! You can now delete {FILENAME_FILTERED_USERS}" + COLOR_ENDC)
        except ValueError as e:
            print(COLOR_FAIL + f"[Migration] Cannot load {FILENAME_FILTERED_USERS}: {e}" + COLOR_ENDC)

    sessions_path = os.path.join(my_username, FILENAME_SESSIONS)
    if os.path.exists(sessions_path):
        try:
            print(f"[Migration] Loading data from {FILENAME_SESSIONS}...")
            sessions_persistent_list = Sessions()
            with open(sessions_path, encoding="utf-8") as json_file:
                sessions = json.load(json_file)
                for session in sessions:
                    session_state = SessionState()
                    session_state.id = session["id"]
                    session_state.args = str(session["args"])
                    session_state.app_version = session.get("app_version", "")
                    session_state.my_username = my_username
                    session_state.my_followers_count = session["profile"].get("followers", 0)
                    session_state.my_following_count = session["profile"].get("following", 0)
                    session_state.totalInteractions = {None: session["total_interactions"]}
                    session_state.successfulInteractions = {None: session["successful_interactions"]}
                    session_state.totalFollowed = {None: session["total_followed"]}
                    session_state.totalScraped = session.get("total_scraped", {})
                    session_state.totalLikes = session["total_likes"]
                    session_state.totalGetProfile = session.get("total_get_profile", 0)
                    session_state.totalUnfollowed = session.get("total_unfollowed", 0)
                    session_state.totalStoriesWatched = session.get("total_stories_watched", 0)
                    session_state.removedMassFollowers = session["removed_mass_followers"]
                    session_state.startTime = datetime.strptime(session["start_time"], '%Y-%m-%d %H:%M:%S.%f')
                    if session.get("finish_time") != "None":
                        session_state.finishTime = datetime.strptime(session["finish_time"], '%Y-%m-%d %H:%M:%S.%f')

                    sessions_persistent_list.append(session_state)
                sessions_persistent_list.persist(my_username)
                print(COLOR_BOLD + f"[Migration] Done! You can now delete {FILENAME_SESSIONS}" + COLOR_ENDC)
        except ValueError as e:
            print(COLOR_FAIL + f"[Migration] Cannot load {FILENAME_SESSIONS}: {e}" + COLOR_ENDC)



from insomniac.database_engine import get_database, add_sessions

FILENAME_SESSIONS = "sessions.json"  # deprecated


class Sessions(list):

    def persist(self, username):
        """
        Save the sessions list to the database and then clear this list.
        """
        database = get_database(username)
        add_sessions(database, self)
        self.clear()



import argparse
import sys

from insomniac.utils import *


def parse_arguments(all_args_dict):
    parser = argparse.ArgumentParser(
        description='Instagram bot for automated Instagram interaction using Android device via ADB',
        add_help=False
    )

    for name, val in all_args_dict.items():
        arg_name = "--{0}".format(name.replace('_', '-'))
        parser.add_argument(arg_name, **val)

    parser.add_argument('--config-file',
                        help='add this argument if you want to load your configuration from a config file. '
                             'Example can be found in config-examples folder')

    if not len(sys.argv) > 1:
        parser.print_help()
        return False, None

    args, unknown_args = parser.parse_known_args()

    if unknown_args:
        print(COLOR_FAIL + "Unknown arguments: " + ", ".join(str(arg) for arg in unknown_args) + COLOR_ENDC)
        parser.print_help()
        return False, None

    if args.config_file is not None:
        if not os.path.exists(args.config_file):
            print(COLOR_FAIL + "Config file {0} could not be found".format(args.config_file) + COLOR_ENDC)
            parser.print_help()
            return False, None

        refresh_args_by_conf_file(args)

    return True, args


def refresh_args_by_conf_file(args):
    if args.config_file is not None:
        if not os.path.exists(args.config_file):
            print(COLOR_FAIL + "Config file {0} could not be found. Continue with previous args".format(args.config_file) + COLOR_ENDC)
            return

        with open(args.config_file, encoding="utf-8") as json_file:
            params = json.load(json_file)

            for param in params:
                if param["enabled"]:
                    args.__setattr__(param["parameter-name"], param["value"])



import base64
import urllib.request
import zlib
from urllib.error import HTTPError

from insomniac.__version__ import __version__
from insomniac.utils import *

HOST = "https://insomniac-bot.com"
PATH_VALIDATE = "/validate"
PATH_ACTIVATE = "/activate/"
PATH_EXTRA_FEATURE = "/extra-feature/"


class ActivationController:
    activation_code = ""
    is_activated = False

    def validate(self, activation_code):
        self.activation_code = activation_code
        if not activation_code == "" and _validate(activation_code):
            self.is_activated = True

        if not self.is_activated:
            dot = '\n    - '
            print_timeless(f"Hi! Since of v3.1.0 all core features in this project are free to use.\n"
                           f"You may want to get more fine grained control over the bot via these features:"
                           f"{dot}{COLOR_BOLD}Filtering{COLOR_ENDC} - skip unwanted accounts by various parameters"
                           f"{dot}{COLOR_BOLD}Scrapping{COLOR_ENDC} - technique that makes interactions "
                           f"significantly safer and faster"
                           f"{dot}{COLOR_BOLD}Working hours{COLOR_ENDC} - the script will wait till specified hours "
                           f"before each session"
                           f"{dot}{COLOR_BOLD}Removing mass followers{COLOR_ENDC} - automate \"cleaning\" you account"
                           f"{dot}{COLOR_BOLD}Analytics tool{COLOR_ENDC} - build presentation that shows your growth\n"
                           f"Activate by supporting our small team: {COLOR_BOLD}{HOST}{PATH_ACTIVATE}{COLOR_ENDC}\n")

    def get_extra_feature(self, module):
        reason = None
        file = None
        try:
            with urllib.request.urlopen(f"{HOST}{PATH_EXTRA_FEATURE}{module}"
                                        f"?activation_code={self.activation_code}"
                                        f"&version={__version__}",
                                        context=ssl.SSLContext()) as response:
                code = response.code
                if code == 200:
                    file = response.read()
        except HTTPError as e:
            code = e.code
            reason = e.reason
        except URLError as e:
            code = -1
            reason = e.reason

        if file is not None:
            return base64.b64decode(zlib.decompress(file))

        if reason is None:
            reason = "Unknown response code"

        print(COLOR_FAIL + f"Cannot get module {module} from v{__version__}: {code} ({reason})" + COLOR_ENDC)
        return None


def print_activation_required_to(action):
    print_timeless(COLOR_FAIL + f"\nActivate the bot to {action}:\n" + COLOR_BOLD + f"{HOST}{PATH_ACTIVATE}" +
                   COLOR_ENDC)


def _validate(activation_code):
    reason = None
    try:
        with urllib.request.urlopen(f"{HOST}{PATH_VALIDATE}?activation_code={activation_code}",
                                    context=ssl.SSLContext()) as response:
            code = response.code
    except HTTPError as e:
        code = e.code
        reason = e.reason
    except URLError as e:
        code = -1
        reason = e.reason

    if code == 200:
        print(COLOR_OKGREEN + "Your activation code is confirmed, welcome!" + COLOR_ENDC)
        return True

    if reason is None:
        reason = "Unknown response code"

    print(COLOR_FAIL + f"Activation code is not confirmed: {code} ({reason})" + COLOR_ENDC)
    return False


class ActivationRequiredException(Exception):
    pass


activation_controller = ActivationController()
